"""Formula 6.28 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot28VplTRdHollowSection(Formula):
    r"""Class representing formula 6.28 for the calculation of [$V_{pl,T,Rd}$]."""

    label = "6.28"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        tau_t_ed: MPA,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
        v_pl_rd: N,
    ) -> None:
        r"""[$V_{pl,T,Rd}$] Calculation of the design plastic shear resistance for structural hollow sections [$N$].

        EN 1993-1-1:2005 art.6.2.7(9) - Formula (6.28)

        Parameters
        ----------
        tau_t_ed : MPA
            [$\tau_{Ed}$] Design shear stress due to St. Venant torsion [$MPa$].
        f_y : MPA
            [$f_y$] Yield strength of the material [$MPa$].
        gamma_m0 : DIMENSIONLESS
            [$\gamma_{M0}$] Partial safety factor for resistance of cross-sections.
        v_pl_rd : N
            [$V_{pl,Rd}$] Plastic shear resistance given in 6.2.6 [$N$].
        """
        super().__init__()
        self.tau_t_ed = tau_t_ed
        self.f_y = f_y
        self.gamma_m0 = gamma_m0
        self.v_pl_rd = v_pl_rd

    @staticmethod
    def _evaluate(
        tau_t_ed: MPA,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
        v_pl_rd: N,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(tau_t_ed=tau_t_ed, v_pl_rd=v_pl_rd)
        raise_if_less_or_equal_to_zero(gamma_m0=gamma_m0, f_y=f_y)

        return (1 - tau_t_ed / ((f_y / np.sqrt(3)) / gamma_m0)) * v_pl_rd

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.28."""
        _equation: str = r"\left (1 - \frac{\tau_{t,Ed}}{\left( f_y / \sqrt{3} \right) / \gamma_{M0}} \right) \cdot V_{pl,Rd}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\tau_{t,Ed}": f"{self.tau_t_ed:.{n}f}",
                r"f_y": f"{self.f_y:.{n}f}",
                r"\gamma_{M0}": f"{self.gamma_m0:.{n}f}",
                r"V_{pl,Rd}": f"{self.v_pl_rd:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"V_{pl,T,Rd}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="N",
        )



"""Testing formula 6.19 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_19 import (
    Form6Dot19CheckDesignElasticShearResistance,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot19CheckDesignElasticShearResistance:
    """Validation for formula 6.19 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        tau_ed = 100.0
        f_y = 355.0
        gamma_m0 = 1.0

        # Object to test
        formula = Form6Dot19CheckDesignElasticShearResistance(tau_ed=tau_ed, f_y=f_y, gamma_m0=gamma_m0)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("tau_ed", "f_y", "gamma_m0"),
        [
            (-100.0, 355.0, 1.0),  # tau_ed is negative
            (100.0, -355.0, 1.0),  # f_y is negative
            (100.0, 355.0, -1.0),  # gamma_m0 is negative
            (100.0, 0.0, 1.0),  # f_y is zero
            (100.0, 355.0, 0.0),  # gamma_m0 is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, tau_ed: float, f_y: float, gamma_m0: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot19CheckDesignElasticShearResistance(tau_ed=tau_ed, f_y=f_y, gamma_m0=gamma_m0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left( \frac{\tau_{Ed}}{f_{y} / (\sqrt{3} \cdot \gamma_{M0})} \leq 1.0 \right) \to "
                r"\left( \frac{100.000}{355.000 / (\sqrt{3} \cdot 1.000)} \leq 1.0 \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        tau_ed = 100.0
        f_y = 355.0
        gamma_m0 = 1.0

        # Object to test
        latex = Form6Dot19CheckDesignElasticShearResistance(tau_ed=tau_ed, f_y=f_y, gamma_m0=gamma_m0).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 5.20 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot20DesignModulusElasticity(Formula):
    """Class representing formula 5.20 for the calculation of the design modulus of elasticity, [$E_{cd}$]."""

    label = "5.20"
    source_document = EN_1992_1_1_2004

    def __init__(self, e_cm: MPA, gamma_ce: DIMENSIONLESS = 1.2) -> None:
        r"""[$E_{cd}$] Design modulus of elasticity.

        EN 1992-1-1:2004 art.5.8.6(3) - Formula (5.20)

        Parameters
        ----------
        e_cm : MPA
            [$E_{cm}$] is the characteristic modulus of elasticity of concrete.
        gamma_ce : DIMENSIONLESS, optional
            [$\gamma_{cE}$] is the factor for the design value of the modulus of elasticity. Default is 1.2 which is the recommended value.
        """
        super().__init__()
        self.e_cm = e_cm
        self.gamma_ce = gamma_ce

    @staticmethod
    def _evaluate(e_cm: MPA, gamma_ce: DIMENSIONLESS) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(gamma_ce=gamma_ce)
        return e_cm / gamma_ce

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.20."""
        return LatexFormula(
            return_symbol=r"E_{cd}",
            result=f"{self:.{n}f}",
            equation=r"\frac{E_{cm}}{\gamma_{CE}}",
            numeric_equation=rf"\frac{{{self.e_cm}}}{{{self.gamma_ce}}}",
            comparison_operator_label="=",
        )



"""Testing formula 5.44 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_44 import Form5Dot44PrestressLoss
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot44PrestressLoss:
    """Validation for formula 5.44 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_p = 1000.0  # mm^2
        e_p = 200000.0  # MPa
        j = [1, 2]  # dimensionless
        delta_sigma_c_t = [10.0, 20.0]  # MPa
        e_cm_t = [30000.0, 30000.0]  # MPa

        # Object to test
        formula = Form5Dot44PrestressLoss(a_p=a_p, e_p=e_p, j=j, delta_sigma_c_t=delta_sigma_c_t, e_cm_t=e_cm_t)

        # Expected result, manually calculated
        manually_calculated_result = 1e6 / 3.0  # N

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a_p", "e_p", "j", "delta_sigma_c_t", "e_cm_t"),
        [
            (-1000.0, 200000.0, [1, 2], [10.0, 20.0], [30000.0, 30000.0]),  # a_p is negative
            (1000.0, -200000.0, [1, 2], [10.0, 20.0], [30000.0, 30000.0]),  # e_p is negative
            (1000.0, 200000.0, [-1, 2], [10.0, 20.0], [30000.0, 30000.0]),  # j contains negative value
            (1000.0, 200000.0, [1, 2], [-10.0, 20.0], [30000.0, 30000.0]),  # delta_sigma_c_t contains negative value
            (1000.0, 200000.0, [1, 2], [10.0, 20.0], [-30000.0, 30000.0]),  # e_cm_t contains negative value
            (1000.0, 200000.0, [1, 2], [10.0, 20.0], [0.0, 30000.0]),  # e_cm_t contains zero value
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a_p: float, e_p: float, j: list, delta_sigma_c_t: list, e_cm_t: list) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot44PrestressLoss(a_p=a_p, e_p=e_p, j=j, delta_sigma_c_t=delta_sigma_c_t, e_cm_t=e_cm_t)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\Delta P_{el} = A_{p} \cdot E_{p} \cdot \sum_{i=1}^{n} \frac{j_{i} \cdot \Delta \sigma_{c,i}(t)}{E_{cm,i}(t)} = "
                r"1000.000 \cdot 200000.000 \cdot \left( \frac{1 \cdot 10.000}{30000.000} + \frac{2 \cdot 20.000}{30000.000} \right) = "
                r"333333.333 \ N",
            ),
            ("short", r"\Delta P_{el} = 333333.333 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_p = 1000.0  # mm^2
        e_p = 200000.0  # MPa
        j = [1, 2]  # dimensionless
        delta_sigma_c_t = [10.0, 20.0]  # MPa
        e_cm_t = [30000.0, 30000.0]  # MPa

        # Object to test
        latex = Form5Dot44PrestressLoss(a_p=a_p, e_p=e_p, j=j, delta_sigma_c_t=delta_sigma_c_t, e_cm_t=e_cm_t).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 6.44 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM, NMM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot44BetaRectangular(Formula):
    r"""Class representing formula 6.44 for the calculation of [$\beta$] for rectangular columns where there are
    eccentricities in both orthogonal directions.
    """

    label = "6.44"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        u1: MM,
        u1_star: MM,
        k: DIMENSIONLESS,
        w_1: NMM,
        e_par: MM,
    ) -> None:
        r"""[$\beta$] Calculation of [$\beta$] where there are eccentricities in both orthogonal directions.

        EN 1992-1-1:2004 art.6.4.3(3) - Formula (6.44)

        Parameters
        ----------
        u1 : MM
            [$u_1$] Basic control perimeter [$mm$].
        u1_star : MM
            [$u_{1^*}$] Reduced basic control perimeter [$mm$].
        k : DIMENSIONLESS
            [$k$] Factor determined from Table 6.1 with the ratio c1/2c2 [$-$].
        w_1 : NMM
            [$W_1$] Calculated for the basic control perimeter [$Nmm$].
        e_par : MM
            [$e_{par}$] Eccentricity parallel to the slab edge [$mm$].
        """
        super().__init__()
        self.u1 = u1
        self.u1_star = u1_star
        self.k = k
        self.w_1 = w_1
        self.e_par = e_par

    @staticmethod
    def _evaluate(
        u1: MM,
        u1_star: MM,
        k: DIMENSIONLESS,
        w_1: NMM,
        e_par: MM,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(u1=u1, k=k, e_par=e_par)
        raise_if_less_or_equal_to_zero(u1_star=u1_star, w_1=w_1)

        return (u1 / u1_star) + k * (u1 / w_1) * e_par

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.44."""
        _equation: str = r"\frac{u_1}{u_{1^*}} + k \cdot \frac{u_1}{W_1} \cdot e_{par}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"u_1": f"{self.u1:.{n}f}",
                r"u_{1^*}": f"{self.u1_star:.{n}f}",
                r"k": f"{self.k:.{n}f}",
                r"W_1": f"{self.w_1:.{n}f}",
                r"e_{par}": f"{self.e_par:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\beta",
            result=f"{self._evaluate(self.u1, self.u1_star, self.k, self.w_1, self.e_par):.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Circular tube cross-section shape."""

from dataclasses import dataclass

from sectionproperties.pre import Geometry
from shapely import Point, Polygon

from blueprints.structural_sections._cross_section import CrossSection
from blueprints.type_alias import MM


@dataclass(frozen=True)
class TubeCrossSection(CrossSection):
    """
    Class to represent a circular tube cross-section using shapely for geometric calculations.

    Parameters
    ----------
    outer_diameter : MM
        The outer diameter of the circular tube cross-section [mm].
    inner_diameter : MM
        The inner diameter of the circular tube cross-section [mm].
    x : MM
        The x-coordinate of the tube's center.
    y : MM
        The y-coordinate of the tube's center.
    name : str
        The name of the rectangular cross-section, default is "Tube".
    """

    outer_diameter: MM
    inner_diameter: MM
    x: MM
    y: MM
    name: str = "Tube"

    def __post_init__(self) -> None:
        """Post-initialization to validate the diameters."""
        if self.outer_diameter <= 0:
            msg = f"Outer diameter must be a positive value, but got {self.outer_diameter}"
            raise ValueError(msg)
        if self.inner_diameter < 0:
            msg = f"Inner diameter cannot be negative, but got {self.inner_diameter}"
            raise ValueError(msg)
        if self.inner_diameter >= self.outer_diameter:
            msg = f"Inner diameter must be smaller than outer diameter, but got inner: {self.inner_diameter}, outer: {self.outer_diameter}"
            raise ValueError(msg)

    @property
    def outer_radius(self) -> MM:
        """
        Calculate the outer radius of the circular tube cross-section [mm].

        Returns
        -------
        MM
            The outer radius of the tube.
        """
        return self.outer_diameter / 2.0

    @property
    def inner_radius(self) -> MM:
        """
        Calculate the inner radius of the circular tube cross-section [mm].

        Returns
        -------
        MM
            The inner radius of the tube.
        """
        return self.inner_diameter / 2.0

    @property
    def wall_thickness(self) -> MM:
        """
        Calculate the thickness of the tube wall [mm].

        Returns
        -------
        MM
            The thickness of the tube wall.
        """
        return self.outer_radius - self.inner_radius

    @property
    def polygon(self) -> Polygon:
        """
        Shapely Polygon representing the circular tube cross-section.

        Returns
        -------
        Polygon
            The shapely Polygon representing the tube.
        """
        center = Point(self.x, self.y)
        quad_segs = 64
        outer_circle = center.buffer(self.outer_radius, quad_segs=quad_segs)
        inner_circle = center.buffer(self.inner_radius, quad_segs=quad_segs)
        difference = outer_circle.difference(inner_circle)
        return Polygon(difference)  # type: ignore[arg-type]

    def geometry(
        self,
        mesh_size: MM | None = None,
    ) -> Geometry:
        """Return the geometry of the tube cross-section.

        Properties
        ----------
        mesh_size : MM
            Maximum mesh element area to be used within
            the Geometry-object finite-element mesh. If not provided, a default value will be used.

        """
        if mesh_size is None:
            minimum_mesh_size = 1.0
            mesh_length = max(self.wall_thickness / 3, minimum_mesh_size)
            mesh_size = mesh_length**2

        tube = Geometry(geom=self.polygon)
        tube.create_mesh(mesh_sizes=mesh_size)
        return tube



"""Testing formula 6.34 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_34 import Form6Dot34CheckAxialForceY
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot34CheckAxialForceY:
    """Validation for formula 6.34 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        n_ed = 50.0
        h_w = 200.0
        t_w = 10.0
        f_y = 355.0
        gamma_m0 = 1.0

        formula = Form6Dot34CheckAxialForceY(n_ed=n_ed, h_w=h_w, t_w=t_w, f_y=f_y, gamma_m0=gamma_m0)

        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("h_w", "t_w", "f_y", "gamma_m0"),
        [
            (-200.0, 10.0, 355.0, 1.0),  # h_w is negative
            (200.0, -10.0, 355.0, 1.0),  # t_w is negative
            (200.0, 10.0, -355.0, 1.0),  # f_y is negative
            (200.0, 10.0, 355.0, -1.0),  # gamma_m0 is negative
            (200.0, 10.0, 355.0, 0.0),  # gamma_m0 is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, h_w: float, t_w: float, f_y: float, gamma_m0: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot34CheckAxialForceY(n_ed=50.0, h_w=h_w, t_w=t_w, f_y=f_y, gamma_m0=gamma_m0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to N_{Ed} \leq \frac{0.5 \cdot h_{w} \cdot t_{w} \cdot f_{y}}{\gamma_{M0}} \to "
                r"50.000 \leq \frac{0.5 \cdot 200.000 \cdot 10.000 \cdot 355.000}{1.000} \to OK",
            ),
            (
                "complete_with_units",
                r"CHECK \to N_{Ed} \leq \frac{0.5 \cdot h_{w} \cdot t_{w} \cdot f_{y}}{\gamma_{M0}} \to "
                r"50.000 \ N \leq \frac{0.5 \cdot 200.000 \ mm \cdot 10.000 \ mm \cdot 355.000 \ MPa}{1.000} \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        n_ed = 50.0
        h_w = 200.0
        t_w = 10.0
        f_y = 355.0
        gamma_m0 = 1.0

        latex = Form6Dot34CheckAxialForceY(n_ed=n_ed, h_w=h_w, t_w=t_w, f_y=f_y, gamma_m0=gamma_m0).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 12.1 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_12_plain_and_lightly_reinforced_concrete_structures.formula_12_1 import (
    Form12Dot1PlainConcreteTensileStrength,
)
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm12Dot1PlainConcreteTensileStrength:
    """Validation for formula 12.1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        alpha_ct = 0.8  # -
        f_ctk_0_05 = 2.5  # MPa
        gamma_c = 1.5  # -

        # Object to test
        form_12_1 = Form12Dot1PlainConcreteTensileStrength(alpha_ct=alpha_ct, f_ctk_0_05=f_ctk_0_05, gamma_c=gamma_c)

        # Expected result, manually calculated
        manually_calculated_result = 1.333  # MPa

        assert round(form_12_1, 3) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("alpha_ct", "f_ctk_0_05", "gamma_c"),
        [
            (-0.8, 2.5, 1.5),
            (0.8, -2.5, 1.5),
        ],
    )
    def test_raise_error_when_negative_values_are_given(
        self,
        alpha_ct: DIMENSIONLESS,
        f_ctk_0_05: DIMENSIONLESS,
        gamma_c: DIMENSIONLESS,
    ) -> None:
        """Test negative values for alpha_ct_pl and f_ctk_0_05."""
        with pytest.raises(NegativeValueError):
            Form12Dot1PlainConcreteTensileStrength(alpha_ct=alpha_ct, f_ctk_0_05=f_ctk_0_05, gamma_c=gamma_c)

    @pytest.mark.parametrize(
        "gamma_c",
        [
            0,
            -1.5,
        ],
    )
    def test_raise_error_when_gamma_c_is_less_or_equal_to_zero(
        self,
        gamma_c: DIMENSIONLESS,
    ) -> None:
        """Test gamma_c less or equal to zero."""
        with pytest.raises(LessOrEqualToZeroError):
            Form12Dot1PlainConcreteTensileStrength(alpha_ct=0.8, f_ctk_0_05=2.5, gamma_c=gamma_c)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"f_{ctd,pl} = \alpha_{ct,pl} \cdot \frac{f_{ctk,0.05}}{\gamma_{C}} = 0.800 \cdot \frac{2.500}{1.500} " r"= 1.333",
            ),
            ("short", r"f_{ctd,pl} = 1.333"),
            (
                "string",
                r"f_{ctd,pl} = \alpha_{ct,pl} \cdot \frac{f_{ctk,0.05}}{\gamma_{C}} = 0.800 \cdot \frac{2.500}{1.500} " r"= 1.333",
            ),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        alpha_ct = 0.8  # -
        f_ctk_0_05 = 2.5  # MPa
        gamma_c = 1.5  # -

        # Object to test
        form_12_1_latex = Form12Dot1PlainConcreteTensileStrength(
            alpha_ct=alpha_ct,
            f_ctk_0_05=f_ctk_0_05,
            gamma_c=gamma_c,
        ).latex()

        actual = {
            "complete": form_12_1_latex.complete,
            "short": form_12_1_latex.short,
            "string": str(form_12_1_latex),
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Contains formula 2.21 from CUR 228."""

from blueprints.codes.cur.cur_228 import CUR_228, R_0
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN_M3, KPA, M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form2Dot21ModulusHorizontalSubgrade(Formula):
    """Representation of equation 2.21 CUR 228."""

    source_document = CUR_228
    label = "2.21"
    n_decimals: int = 2

    def __init__(self, r: M, e_p: KPA, alpha: DIMENSIONLESS) -> None:
        """Calculates the modulus of horizontal subgrade reaction (k_h) using Menard stiffness for r >= 0.3 m.

        Parameters
        ----------
        r: M
            The radius of a foundation pile [m]:
            r >= 0.3 m
        e_p: KPA
            Elastic modulus of Ménard [kPa]:
            e_p ≈ beta * q_c
            beta: DIMENSIONLESS
                Dependent on soil type [-]: (table 2.1)
            q_c: KPA
                Cone resistance [kPa]
        alpha: DIMENSIONLESS
            Factor dependent on soil type [-]: (table 2.1)
        """
        super().__init__()
        self.r = float(r)
        self.e_p = float(e_p)
        self.alpha = float(alpha)

    @staticmethod
    def _evaluate(r: M, e_p: KPA, alpha: DIMENSIONLESS) -> KN_M3:
        """Evaluates the formula, for more information see the __init__ method."""
        r_0 = R_0
        raise_if_negative(e_p=e_p)
        raise_if_less_or_equal_to_zero(r=r, alpha=alpha)

        if r >= r_0:
            return 3 * e_p / (1.3 * r_0 * (2.65 * r / r_0) ** alpha + alpha * r)
        msg = "Radius is smaller than 0.3m, use: Eq2Dot21ModulusHorizontalSubgrade"
        raise ValueError(msg)

    def latex(self, n: int = 3) -> LatexFormula:
        """Latex representation of the full equation including result.

        Parameters
        ----------
        n: int
            Number of decimals to round the result to

        Returns
        -------
        LatexFormula
            Latex representation of the equation

        """
        n = self.n_decimals

        return LatexFormula(
            return_symbol="k_{h}",
            equation=r"\frac{1}{3 \cdot E_{p}} \cdot "
            r"\left[1.3 \cdot R_{0} "
            r"\left( 2.65 \frac{R}{R_0}\right)^\alpha"
            r" + \alpha \cdot  R \right]",
            numeric_equation=rf"\frac{{1}}{{3 \cdot {self.e_p:.{n}}}} \cdot"
            rf"\left[1.3 \cdot {R_0:.{n}} "
            rf"\left( 2.65 \cdot \frac{{{self.r:.{n}}}}{{{R_0:.{n}}}}\right)^{{{self.alpha:.{n}f}}}"
            rf"+ {self.alpha:.{n}} \cdot {self.r:.{n}}\right]",
            result=f"{self:.{n}f}",
            unit="kN/m^3",
        )



"""Testing formula 9.13 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_9_detailling_and_specific_rules.formula_9_13 import Form9Dot13TensileForceToBeAnchored
from blueprints.validations import NegativeValueError


class TestForm9Dot13TensileForceToBeAnchored:
    """Validation for formula 9.13 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        r = 100  # mm
        z_e = 50  # mm
        z_i = 20  # mm
        form_9_13 = Form9Dot13TensileForceToBeAnchored(r=r, z_e=z_e, z_i=z_i)

        # Expected result, manually calculated
        manually_calculated_result = 250

        assert form_9_13 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_r_is_given(self) -> None:
        """Test whether negative value error is raised if r is negative."""
        r = -100  # mm
        z_e = 50  # mm
        z_i = 20  # mm

        with pytest.raises(NegativeValueError):
            Form9Dot13TensileForceToBeAnchored(r=r, z_e=z_e, z_i=z_i)

    def test_raise_error_when_negative_z_e_is_given(self) -> None:
        """Test whether negative value error is raised if z_e is negative."""
        r = 100  # mm
        z_e = -50  # mm
        z_i = 20  # mm

        with pytest.raises(NegativeValueError):
            Form9Dot13TensileForceToBeAnchored(r=r, z_e=z_e, z_i=z_i)

    def test_raise_error_when_negative_z_i_is_given(self) -> None:
        """Test whether negative value error is raised if z_i is negative."""
        r = 100  # mm
        z_e = 50  # mm
        z_i = -20  # mm

        with pytest.raises(NegativeValueError):
            Form9Dot13TensileForceToBeAnchored(r=r, z_e=z_e, z_i=z_i)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"F_s = R \cdot z_e / z_i = 100.00 \cdot 50.00 / 20.00 = 250.00",
            ),
            ("short", r"F_s = 250.00"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        r = 100  # mm
        z_e = 50  # mm
        z_i = 20  # mm

        # Object to test
        form_9_13_latex = Form9Dot13TensileForceToBeAnchored(r=r, z_e=z_e, z_i=z_i).latex()

        actual = {"complete": form_9_13_latex.complete, "short": form_9_13_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 7.5 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form7Dot5AdjustedBondStrengthRatio(Formula):
    r"""Class representing formula 7.5 for the calculation of [$\xi_1$]."""

    label = "7.5"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        xi: DIMENSIONLESS,
        diam_s: MM,
        diam_p: MM,
    ) -> None:
        r"""[$\xi_1$] Calculation of the ratio of bond strength of presetressing and reinforcing steel [$-$].
        Note: if only prestressing steel is used to control cracking: [$\xi_1 = \sqrt{\xi}$].

        EN 1992-1-1:2004 art.7.3.2(3) - Formula (7.5)

        Parameters
        ----------
        xi : DIMENSIONLESS
            [$\xi$] Ratio of bond strength of prestressing and reinforcing steel.
        diam_s : MM
            [$⌀_s$] Largest bar diameter of reinforcing steel [$mm$].
        diam_p : MM
            [$⌀_p$] Equivalent diameter of tendon [$mm$].
        """
        super().__init__()
        self.xi = xi
        self.diam_s = diam_s
        self.diam_p = diam_p

    @staticmethod
    def _evaluate(
        xi: DIMENSIONLESS,
        diam_s: MM,
        diam_p: MM,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(xi=xi, diam_s=diam_s)
        raise_if_less_or_equal_to_zero(diam_p=diam_p)

        return np.sqrt(xi * (diam_s / diam_p))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.5."""
        _equation: str = r"\sqrt{\xi \cdot \left( \frac{⌀_s}{⌀_p} \right)}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\xi": f"{self.xi:.{n}f}",
                r"⌀_s": f"{self.diam_s:.{n}f}",
                r"⌀_p": f"{self.diam_p:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\xi_1",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Formula 5.7 from EN 1993-5:2007 Chapter 5 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1993_5_2007 import EN_1993_5_2007
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_fraction
from blueprints.type_alias import DIMENSIONLESS, KN, MM, MPA
from blueprints.unit_conversion import N_TO_KN
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot7ShearBucklingResistance(Formula):
    """Class representing formula 5.7 for shear buckling resistance."""

    label = "5.7"
    source_document = EN_1993_5_2007

    def __init__(
        self,
        h: MM,
        t_f: MM,
        t_w: MM,
        f_bv: MPA,
        gamma_m_0: DIMENSIONLESS,
    ) -> None:
        r"""[$V_{b,Rd}$] Calculate the shear buckling resistance [$kN$].

        EN 1993-5:2007(E) art.5.2.2(7) - Formula (5.7)

        Parameters
        ----------
        h : MM
            [$h$] Height of the web in [$mm$].
        t_f : MM
            [$t_{f}$] Thickness of the flange in [$mm$].
        t_w : MM
            [$t_{w}$] Thickness of the web in [$mm$].
        f_bv : MPA
            [$f_{bv}$] Shear buckling strength according to Table 6-1 of EN 1993-1-3 for a web without stiffening
            at the support and for a relative web slenderness [$MPa$].
        gamma_m_0 : DIMENSIONLESS
            [$\gamma_{M0}$] Partial factor for material properties [$-$].
        """
        super().__init__()
        self.h: float = h
        self.t_f: float = t_f
        self.t_w: float = t_w
        self.f_bv: float = f_bv
        self.gamma_m_0: float = gamma_m_0

    @staticmethod
    def _evaluate(
        h: MM,
        t_f: MM,
        t_w: MM,
        f_bv: MPA,
        gamma_m_0: DIMENSIONLESS,
    ) -> KN:
        """Evaluates the formula for shear buckling resistance."""
        raise_if_less_or_equal_to_zero(
            h=h,
            t_f=t_f,
            t_w=t_w,
            f_bv=f_bv,
            gamma_m_0=gamma_m_0,
        )
        if t_f >= h:
            raise ValueError("The thickness of the flange should be less than the height of the web.")
        return ((h - t_f) * t_w * f_bv / gamma_m_0) * N_TO_KN

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 5.7."""
        return LatexFormula(
            return_symbol=r"V_{b,Rd}",
            result=f"{self:.{n}f}",
            equation=latex_fraction(numerator=r"\left(h - t_f \right) t_w f_{bv}", denominator=r"\gamma_{M0}"),
            numeric_equation=latex_fraction(
                numerator=rf"({self.h:.{n}f} - {self.t_f:.{n}f}) \cdot {self.t_w} \cdot {self.f_bv:.{n}f}",
                denominator=self.gamma_m_0,
            ),
            comparison_operator_label="=",
        )



"""Testing subformulas a to g from 6.18 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_18_sub_av import (
    Form6Dot18SubARolledIandHSection,
    Form6Dot18SubBRolledChannelSection,
    Form6Dot18SubCTSectionRolled,
    Form6Dot18SubCTSectionWelded,
    Form6Dot18SubDWeldedIHandBoxSection,
    Form6Dot18SubEWeldedIHandBoxSection,
    Form6Dot18SubF1RolledRectangularHollowSectionDepth,
    Form6Dot18SubF2RolledRectangularHollowSectionWidth,
    Form6Dot18SubGCircularHollowSection,
)
from blueprints.validations import ListsNotSameLengthError, NegativeValueError


class TestForm6Dot18SubARolledIandHSection:
    """Validation for formula 6.18suba from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 10000.0
        b = 200.0
        hw = 250.0
        r = 10.0
        tf = 15.0
        tw = 8.0
        eta = 1.0

        formula = Form6Dot18SubARolledIandHSection(a=a, b=b, hw=hw, r=r, tf=tf, tw=tw, eta=eta)
        manually_calculated_result = 4420.0  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "b", "hw", "r", "tf", "tw", "eta"),
        [
            (-10000.0, 200.0, 250.0, 10.0, 15.0, 8.0, 1.0),  # a is negative
            (10000.0, -200.0, 250.0, 10.0, 15.0, 8.0, 1.0),  # b is negative
            (10000.0, 200.0, -250.0, 10.0, 15.0, 8.0, 1.0),  # hw is negative
            (10000.0, 200.0, 250.0, -10.0, 15.0, 8.0, 1.0),  # r is negative
            (10000.0, 200.0, 250.0, 10.0, -15.0, 8.0, 1.0),  # tf is negative
            (10000.0, 200.0, 250.0, 10.0, 15.0, -8.0, 1.0),  # tw is negative
            (10000.0, 200.0, 250.0, 10.0, 15.0, 8.0, -1.0),  # eta is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, b: float, hw: float, r: float, tf: float, tw: float, eta: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot18SubARolledIandHSection(a=a, b=b, hw=hw, r=r, tf=tf, tw=tw, eta=eta)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = max(A - 2 \cdot b \cdot t_f + (t_w + 2 \cdot r) \cdot t_f; \eta \cdot h_w \cdot t_w) = "
                r"max(10000.000 - 2 \cdot 200.000 \cdot 15.000 + (8.000 + 2 \cdot 10.000) \cdot 15.000; 1.000 \cdot 250.000 \cdot 8.000) = "
                r"4420.000 \ mm^2",
            ),
            ("short", r"A_v = 4420.000 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 10000.0
        b = 200.0
        hw = 250.0
        r = 10.0
        tf = 15.0
        tw = 8.0
        eta = 1.0

        latex = Form6Dot18SubARolledIandHSection(a=a, b=b, hw=hw, r=r, tf=tf, tw=tw, eta=eta).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot18SubBRolledChannelSection:
    """Validation for formula 6.18subb from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 8000.0
        b = 150.0
        tf = 12.0
        tw = 6.0
        r = 8.0

        formula = Form6Dot18SubBRolledChannelSection(a=a, b=b, tf=tf, tw=tw, r=r)
        manually_calculated_result = 4568.0  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "b", "tf", "tw", "r"),
        [
            (-8000.0, 150.0, 12.0, 6.0, 8.0),  # a is negative
            (8000.0, -150.0, 12.0, 6.0, 8.0),  # b is negative
            (8000.0, 150.0, -12.0, 6.0, 8.0),  # tf is negative
            (8000.0, 150.0, 12.0, -6.0, 8.0),  # tw is negative
            (8000.0, 150.0, 12.0, 6.0, -8.0),  # r is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, b: float, tf: float, tw: float, r: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot18SubBRolledChannelSection(a=a, b=b, tf=tf, tw=tw, r=r)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = A - 2 \cdot b \cdot t_f + (t_w + r) \cdot t_f = "
                r"8000.000 - 2 \cdot 150.000 \cdot 12.000 + (6.000 + 8.000) \cdot 12.000 = 4568.000 \ mm^2",
            ),
            ("short", r"A_v = 4568.000 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 8000.0
        b = 150.0
        tf = 12.0
        tw = 6.0
        r = 8.0

        latex = Form6Dot18SubBRolledChannelSection(a=a, b=b, tf=tf, tw=tw, r=r).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot18SubCRolledTSectionRolled:
    """Validation for formula 6.18subc from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 6000.0
        b = 100.0
        tf = 10.0
        tw = 8.0
        r = 5.0

        formula = Form6Dot18SubCTSectionRolled(a=a, b=b, tf=tf, tw=tw, r=r)
        manually_calculated_result = 5090.0  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "b", "tf", "tw", "r"),
        [
            (-6000.0, 100.0, 10.0, 8.0, 5.0),  # a is negative
            (6000.0, -100.0, 10.0, 8.0, 5.0),  # b is negative
            (6000.0, 100.0, -10.0, 8.0, 5.0),  # tf is negative
            (6000.0, 100.0, 10.0, -8.0, 5.0),  # tw is negative
            (6000.0, 100.0, 10.0, 8.0, -5.0),  # r is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, b: float, tf: float, tw: float, r: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot18SubCTSectionRolled(a=a, b=b, tf=tf, tw=tw, r=r)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = A - b \cdot t_f + (t_w + 2 \cdot r) \cdot \frac{t_f}{2} = "
                r"6000.000 - 100.000 \cdot 10.000 + (8.000 + 2 \cdot 5.000) \cdot \frac{10.000}{2} = 5090.000 \ mm^2",
            ),
            ("short", r"A_v = 5090.000 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 6000.0
        b = 100.0
        tf = 10.0
        tw = 8.0
        r = 5.0

        latex = Form6Dot18SubCTSectionRolled(a=a, b=b, tf=tf, tw=tw, r=r).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot18SubCRolledTSectionWelded:
    """Validation for formula 6.18subc from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        tf = 10.0
        tw = 8.0
        h = 200.0

        formula = Form6Dot18SubCTSectionWelded(tf=tf, tw=tw, h=h)
        manually_calculated_result = 8000.0  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("tf", "tw", "h"),
        [
            (-10.0, 8.0, 200.0),  # tf is negative
            (10.0, -8.0, 200.0),  # tw is negative
            (10.0, 8.0, -200.0),  # h is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, tf: float, tw: float, h: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot18SubCTSectionWelded(tf=tf, tw=tw, h=h)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = t_w \cdot (h \cdot t_f / 2) = "
                r"8.000 \cdot (200.000 \cdot 10.000 / 2) = 8000.000 \ mm^2",
            ),
            ("short", r"A_v = 8000.000 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        tf = 10.0
        tw = 8.0
        h = 200.0

        latex = Form6Dot18SubCTSectionWelded(tf=tf, tw=tw, h=h).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot18SubDWeldedIHandBoxSection:
    """Validation for formula 6.18subd from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        hw_list = [250.0, 300.0]
        tw_list = [8.0, 10.0]
        eta = 1.0

        formula = Form6Dot18SubDWeldedIHandBoxSection(hw_list=hw_list, tw_list=tw_list, eta=eta)
        manually_calculated_result = 5000.0  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("hw_list", "tw_list", "eta"),
        [
            ([250.0, -300.0], [8.0, 10.0], 1.0),  # hw_list contains negative value
            ([250.0, 300.0], [8.0, -10.0], 1.0),  # tw_list contains negative value
            ([250.0, 300.0], [8.0, 10.0], -1.0),  # eta is negative
            ([250.0], [8.0, 10.0], 1.0),  # hw_list and tw_list are not the same length
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, hw_list: list[float], tw_list: list[float], eta: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, ListsNotSameLengthError)):
            Form6Dot18SubDWeldedIHandBoxSection(hw_list=hw_list, tw_list=tw_list, eta=eta)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = \eta \cdot \sum (h_{w} \cdot t_{w}) = "
                r"1.000 \cdot (250.000 \cdot 8.000 + 300.000 \cdot 10.000) = 5000.000 \ mm^2",
            ),
            ("short", r"A_v = 5000.000 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        hw_list = [250.0, 300.0]
        tw_list = [8.0, 10.0]
        eta = 1.0

        latex = Form6Dot18SubDWeldedIHandBoxSection(hw_list=hw_list, tw_list=tw_list, eta=eta).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot18SubEWeldedIHandBoxSection:
    """Validation for formula 6.18sube from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 12000.0
        hw_list = [250.0, 300.0]
        tw_list = [8.0, 10.0]

        formula = Form6Dot18SubEWeldedIHandBoxSection(a=a, hw_list=hw_list, tw_list=tw_list)
        manually_calculated_result = 7000.0  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "hw_list", "tw_list"),
        [
            (-12000.0, [250.0, 300.0], [8.0, 10.0]),  # a is negative
            (12000.0, [250.0, -300.0], [8.0, 10.0]),  # hw_list contains negative value
            (12000.0, [250.0, 300.0], [8.0, -10.0]),  # tw_list contains negative value
            (12000.0, [250.0, 300.0], [8.0, 10.0, 12.0]),  # hw_list and tw_list are not the same length
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, hw_list: list[float], tw_list: list[float]) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, ListsNotSameLengthError)):
            Form6Dot18SubEWeldedIHandBoxSection(a=a, hw_list=hw_list, tw_list=tw_list)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = A - \sum (h_{w} \cdot t_{w}) = "
                r"12000.000 - (250.000 \cdot 8.000 + 300.000 \cdot 10.000) = 7000.000 \ mm^2",
            ),
            ("short", r"A_v = 7000.000 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 12000.0
        hw_list = [250.0, 300.0]
        tw_list = [8.0, 10.0]

        latex = Form6Dot18SubEWeldedIHandBoxSection(a=a, hw_list=hw_list, tw_list=tw_list).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot18SubF1RolledRectangularHollowSectionDepth:
    """Validation for formula 6.18subf1 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 5000.0
        b = 100.0
        h = 200.0

        formula = Form6Dot18SubF1RolledRectangularHollowSectionDepth(a=a, b=b, h=h)
        manually_calculated_result = 3333.333  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "b", "h"),
        [
            (-5000.0, 100.0, 200.0),  # a is negative
            (5000.0, -100.0, 200.0),  # b is negative
            (5000.0, 100.0, -200.0),  # h is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, b: float, h: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot18SubF1RolledRectangularHollowSectionDepth(a=a, b=b, h=h)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = \frac{A \cdot h}{b + h} = "
                r"\frac{5000.000 \cdot 200.000}{100.000 + 200.000} = 3333.333 \ mm^2",
            ),
            ("short", r"A_v = 3333.333 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 5000.0
        b = 100.0
        h = 200.0

        latex = Form6Dot18SubF1RolledRectangularHollowSectionDepth(a=a, b=b, h=h).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot18SubF2RolledRectangularHollowSectionWidth:
    """Validation for formula 6.18subf2 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 5000.0
        b = 100.0
        h = 200.0

        formula = Form6Dot18SubF2RolledRectangularHollowSectionWidth(a=a, b=b, h=h)
        manually_calculated_result = 1666.667  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "b", "h"),
        [
            (-5000.0, 100.0, 200.0),  # a is negative
            (5000.0, -100.0, 200.0),  # b is negative
            (5000.0, 100.0, -200.0),  # h is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, b: float, h: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot18SubF2RolledRectangularHollowSectionWidth(a=a, b=b, h=h)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = \frac{A \cdot b}{b + h} = "
                r"\frac{5000.000 \cdot 100.000}{100.000 + 200.000} = 1666.667 \ mm^2",
            ),
            ("short", r"A_v = 1666.667 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 5000.0
        b = 100.0
        h = 200.0

        latex = Form6Dot18SubF2RolledRectangularHollowSectionWidth(a=a, b=b, h=h).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot18SubGCircularHollowSection:
    """Validation for formula 6.18subg from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 4000.0

        formula = Form6Dot18SubGCircularHollowSection(a=a)
        manually_calculated_result = 2546.4790899483937  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        "a",
        [
            -4000.0,  # a is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot18SubGCircularHollowSection(a=a)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_v = \frac{2 \cdot A}{\pi} = "
                r"\frac{2 \cdot 4000.000}{\pi} = 2546.479 \ mm^2",
            ),
            ("short", r"A_v = 2546.479 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 4000.0

        latex = Form6Dot18SubGCircularHollowSection(a=a).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Tests for the circular cornered cross section."""

import pytest

from blueprints.structural_sections.cross_section_cornered import CircularCorneredCrossSection
from blueprints.validations import NegativeValueError


class TestCircularCorneredCrossSection:
    """Tests for the CircularCorneredCrossSection class."""

    def test_polygon(self, qcs_cross_section: CircularCorneredCrossSection) -> None:
        """Test the polygon property."""
        polygon = qcs_cross_section.polygon
        assert polygon.is_valid
        assert len(polygon.exterior.coords) > 0

    def test_geometry(self, qcs_cross_section: CircularCorneredCrossSection) -> None:
        """Test the geometry property."""
        geometry = qcs_cross_section.geometry()
        assert geometry is not None

    @pytest.mark.parametrize(
        "kwargs",
        [
            {"thickness_vertical": -1, "thickness_horizontal": 10, "inner_radius": 5, "outer_radius": 10},
            {"thickness_vertical": 10, "thickness_horizontal": -1, "inner_radius": 5, "outer_radius": 10},
            {"thickness_vertical": 10, "thickness_horizontal": 10, "inner_radius": -1, "outer_radius": 10},
            {"thickness_vertical": 10, "thickness_horizontal": 10, "inner_radius": 5, "outer_radius": -1},
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, kwargs: dict) -> None:
        """Test NegativeValueError is raised for negative values."""
        with pytest.raises(NegativeValueError):
            CircularCorneredCrossSection(**kwargs)

    def test_invalid_outer_radius_greater_than_inner_plus_thickness(self) -> None:
        """Test initialization with an outer radius greater than inner radius plus thickness."""
        with pytest.raises(
            ValueError,
            match="Outer radius 20 must be smaller than or equal to inner radius 5 plus the thickness 10",
        ):
            CircularCorneredCrossSection(
                thickness_vertical=10,
                thickness_horizontal=10,
                inner_radius=5,
                outer_radius=20,
            )

    def test_invalid_corner_direction(self) -> None:
        """Test initialization with an invalid corner direction."""
        with pytest.raises(ValueError, match="corner_direction must be one of 0, 1, 2, or 3, got 4"):
            CircularCorneredCrossSection(
                thickness_vertical=10,
                thickness_horizontal=10,
                inner_radius=5,
                outer_radius=10,
                corner_direction=4,
            )



"""Module for concrete material properties."""

from dataclasses import dataclass, field
from enum import Enum

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_3_materials.table_3_1 import (
    SteelStrengthClass,
    Table3Dot1NominalValuesHotRolledStructuralSteel,
)
from blueprints.type_alias import DIMENSIONLESS, KG_M3, MM, MPA, PER_DEGREE

# default data from NEN-EN 1992-1-1 3.1.1(1)
STEEL_YOUNG_MODULUS = 210_000.0  # [MPa]
STEEL_POISSON_RATIO = 0.3  # [-]
STEEL_THERMAL_COEFFICIENT = 1.2e-5  # [1/°C]


class DiagramType(Enum):
    """Enumeration of diagram types of stress-strain relations."""

    BI_LINEAR = "Bi-Linear"
    PARABOLIC = "Parabolic"


@dataclass(frozen=True)
class SteelMaterial:
    r"""Representation of the strength and deformation characteristics for steel material.

    Parameters
    ----------
    steel_class: SteelStrengthClass
        Enumeration of steel strength classes (default: S355)
    density: KG_M3
        Unit mass of steel [$kg/m^3$] (default= 7850.0)
    diagram_type: DiagramType
        Type of stress-strain diagram (default= Bi-Linear)
    quality_class: str
        Quality class of the steel material (default= None)
    custom_name: str
        Use a custom name for the steel material (default= steel class name)
    custom_e_modulus: MPA
        Use a custom modulus of elasticity of steel [$MPa$]
        If no custom value is given, the value [$E$] from standard tables is used.
    custom_poisson_ratio: DIMENSIONLESS
        Use a custom Poisson's ratio for the steel material (default= 0.3)
    custom_thermal_coefficient: PER_DEGREE
        Use a custom thermal coefficient for the steel material (default= 1.2e-5)
    custom_yield_strength: MPA
        Use a custom yield strength for the steel material (default= None)
    custom_ultimate_strength: MPA
        Use a custom ultimate strength for the steel material (default= None)

    """

    steel_class: SteelStrengthClass = field(default=SteelStrengthClass.S355)
    density: KG_M3 = field(default=7850.0, metadata={"unit": "kg/m³"})
    diagram_type: DiagramType = field(default=DiagramType.BI_LINEAR)
    quality_class: str | None = field(default=None)
    custom_name: str | None = field(default=None, compare=False)
    custom_e_modulus: MPA | None = field(default=None, metadata={"unit": "MPa"})
    custom_poisson_ratio: DIMENSIONLESS | None = field(default=None)
    custom_thermal_coefficient: PER_DEGREE | None = field(default=None)
    custom_yield_strength: MPA | None = field(default=None)
    custom_ultimate_strength: MPA | None = field(default=None)

    @property
    def name(self) -> str:
        """Name of the steel material.

        Returns
        -------
        str
            Example: "S355"
        """
        if self.custom_name:
            return self.custom_name
        standard_group = self.steel_class.value[0].value
        steel_class = self.steel_class.value[1]
        return f"{steel_class} ({standard_group})"

    @property
    def e_modulus(self) -> MPA:
        """Modulus of elasticity of the steel material.

        Returns
        -------
        MPA
            Modulus of elasticity of steel [$MPa$]
        """
        if self.custom_e_modulus:
            return self.custom_e_modulus
        return STEEL_YOUNG_MODULUS

    @property
    def poisson_ratio(self) -> DIMENSIONLESS:
        """Poisson's ratio of the steel material.

        Returns
        -------
        DIMENSIONLESS
            Poisson's ratio of the material
        """
        if self.custom_poisson_ratio:
            return self.custom_poisson_ratio
        return STEEL_POISSON_RATIO

    @property
    def thermal_coefficient(self) -> PER_DEGREE:
        """Thermal coefficient of the steel material [1/°C].

        Returns
        -------
        PER_DEGREE
            Thermal coefficient of the material
        """
        if self.custom_thermal_coefficient:
            return self.custom_thermal_coefficient
        return STEEL_THERMAL_COEFFICIENT

    @property
    def shear_modulus(self) -> MPA:
        """Shear modulus of the steel material.

        Returns
        -------
        MPA
            Shear modulus of the material
        """
        return self.e_modulus / (2 * (1 + self.poisson_ratio))

    def yield_strength(self, thickness: MM) -> MPA | None:
        """Yield strength of the steel material for steel [$f_y$].

        Parameters
        ----------
        thickness: MM
            Nominal thickness of the steel element [$mm$]

        Returns
        -------
        MPA | None
            Yield strength of the material at the given temperature
        """
        if self.custom_yield_strength:
            return self.custom_yield_strength
        return Table3Dot1NominalValuesHotRolledStructuralSteel(steel_class=self.steel_class, thickness=thickness).fy

    def ultimate_strength(self, thickness: MM) -> MPA | None:
        """Ultimate strength of the steel material for steel [$f_u$].

        Parameters
        ----------
        thickness: MM
            Nominal thickness of the steel element [$mm$]

        Returns
        -------
        MPA | None
            Ultimate strength of the material at the given temperature
        """
        if self.custom_ultimate_strength:
            return self.custom_ultimate_strength
        return Table3Dot1NominalValuesHotRolledStructuralSteel(steel_class=self.steel_class, thickness=thickness).fu



"""Module contains tests for formula 2.1a from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""

import pytest

from blueprints.codes.eurocode.nen_9997_1_c2_2017.chapter_2_basic_of_geotechnical_design.formula_2_1_a import Form2Dot1aDesignValueLoad
from blueprints.codes.eurocode.nen_9997_1_c2_2017.chapter_2_basic_of_geotechnical_design.formula_2_1_b import Form2Dot1bRepresentativeValue
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import NegativeValueError


class TestForm2Dot1aDesignValueLoad:
    """Class containing tests for formula 2.1a from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""

    @pytest.mark.parametrize(
        ("gamma_f", "f_rep", "expected_result"),
        [
            (1.35, 100, 135),
            (1.35, 0.0, 0.0),
            (1.35, -100, -135),
        ],
    )
    def test_formula_2_1_a(
        self,
        gamma_f: DIMENSIONLESS,
        f_rep: float,
        expected_result: float,
    ) -> None:
        """Method to test formula 2.1a from NEN 9997-1+C2:2017: Chapter 2: Basis of geotechnical design."""
        assert Form2Dot1aDesignValueLoad(gamma_f=gamma_f, f_rep=f_rep) == pytest.approx(expected_result, rel=1e-9)

    def test_raise_error_if_gamma_f_is_negative(self) -> None:
        """Test if an error is raised when gamma_f is negative."""
        with pytest.raises(NegativeValueError):
            Form2Dot1aDesignValueLoad(gamma_f=-1.35, f_rep=100)

    def test_integration_with_2_1_b(self) -> None:
        """Test the integration of formula 2.1a with 2.1b."""
        f_rep = Form2Dot1bRepresentativeValue(psi=2, f_k=50)
        gamma_f = 1.35
        result = Form2Dot1aDesignValueLoad(gamma_f=gamma_f, f_rep=f_rep)

        assert result == pytest.approx(expected=135, rel=1e-9)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"F_d = \gamma_F \cdot F_{rep} = 1.20 \cdot 100.00 = 120.00",
            ),
            ("short", r"F_d = 120.00"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        gamma_f = 1.2  # [-]
        f_rep = 100  # kN

        # Object to test
        form_2_1_a_latex = Form2Dot1aDesignValueLoad(gamma_f=gamma_f, f_rep=f_rep).latex()

        actual = {"complete": form_2_1_a_latex.complete, "short": form_2_1_a_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 5.16 from EN 1993-5:2007 Chapter 5 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1993_5_2007 import EN_1993_5_2007
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM2, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot16PlasticDesignResistance(Formula):
    """Class representing formula 5.16 for the calculation of the plastic design resistance of the cross-section, [$N_{pl,Rd}$]."""

    label = "5.16"
    source_document = EN_1993_5_2007

    def __init__(
        self,
        a: MM2,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
    ) -> None:
        r"""[$N_{pl,Rd}$] Plastic design resistance of the cross-section [$N$].

        EN 1993-5:2007 art.5.2.3 (9) - Formula (5.16)

        Parameters
        ----------
        a : MM2
            [$A$] Area of the cross-section [$mm^2$].
        f_y : MPA
            [$f_y$] Yield strength [$MPa$].
        gamma_m0 : DIMENSIONLESS
            [$\gamma_{M0}$] Partial factor according to 5.1.1 (4) [$-$].
        """
        super().__init__()
        self.a = a
        self.f_y = f_y
        self.gamma_m0 = gamma_m0

    @staticmethod
    def _evaluate(
        a: MM2,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(gamma_m0=gamma_m0)
        raise_if_negative(f_y=f_y, a=a)

        return (a * f_y) / gamma_m0

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.16."""
        return LatexFormula(
            return_symbol=r"N_{pl,Rd}",
            result=f"{self:.{n}f}",
            equation=r"\frac{A \cdot f_y}{\gamma_{M0}}",
            numeric_equation=rf"\frac{{{self.a:.{n}f} \cdot {self.f_y:.{n}f}}}{{{self.gamma_m0:.{n}f}}}",
            comparison_operator_label="=",
            unit="N",
        )



"""Testing formula 6.5 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_5 import Form6Dot5ShearForceCheck
from blueprints.validations import NegativeValueError


class TestForm6Dot5ShearForceCheck:
    """Validation for formula 6.5 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed = 1.4e6  # N
        b_w = 300.0  # MM
        d = 500.0  # MM
        nu = 0.6  # DIMENSIONLESS
        f_cd = 30.0  # MPA

        # Object to test
        formula = Form6Dot5ShearForceCheck(v_ed=v_ed, b_w=b_w, d=d, nu=nu, f_cd=f_cd)

        # Expected result, manually calculated
        expected_result = False

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("v_ed", "b_w", "d", "nu", "f_cd"),
        [
            (-100.0, 300.0, 500.0, 0.6, 30.0),  # v_ed is negative
            (100.0, -300.0, 500.0, 0.6, 30.0),  # b_w is negative
            (100.0, 300.0, -500.0, 0.6, 30.0),  # d is negative
            (100.0, 300.0, 500.0, -0.6, 30.0),  # nu is negative
            (100.0, 300.0, 500.0, 0.6, -30.0),  # f_cd is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed: float, b_w: float, d: float, nu: float, f_cd: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot5ShearForceCheck(v_ed=v_ed, b_w=b_w, d=d, nu=nu, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to V_{Ed} \leq 0.5 \cdot b_w \cdot d \cdot \nu \cdot f_{cd} \to 1400000.000 \leq "
                r"0.5 \cdot 300.000 \cdot 500.000 \cdot 0.600 \cdot 30.000 \to \text{Not OK}",
            ),
            ("short", r"CHECK \to \text{Not OK}"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed = 1.4e6  # N
        b_w = 300.0  # MM
        d = 500.0  # MM
        nu = 0.6  # DIMENSIONLESS
        f_cd = 30.0  # MPA

        # Object to test
        latex = Form6Dot5ShearForceCheck(v_ed=v_ed, b_w=b_w, d=d, nu=nu, f_cd=f_cd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.7 of NEN-EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_7 import (
    Form6Dot7DesignUltimateResistanceNetCrossSection,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot7DesignUltimateResistanceNetCrossSection:
    """Validation for formula 6.7 from NEN-EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_net = 2000.0
        f_u = 400.0
        gamma_m2 = 1.1

        # Object to test
        formula = Form6Dot7DesignUltimateResistanceNetCrossSection(a_net=a_net, f_u=f_u, gamma_m2=gamma_m2)

        # Expected result, manually calculated
        manually_calculated_result = 654545.454545454  # N

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a_net", "f_u", "gamma_m2"),
        [
            (2000.0, 400.0, 0.0),  # gamma_m2 is zero
            (-2000.0, 400.0, 1.1),  # a_net is negative
            (2000.0, -400.0, 1.1),  # f_u is negative
            (2000.0, 400.0, -1.1),  # gamma_m2 is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a_net: float, f_u: float, gamma_m2: float) -> None:
        """Test invalid values."""
        with pytest.raises((LessOrEqualToZeroError, NegativeValueError)):
            Form6Dot7DesignUltimateResistanceNetCrossSection(a_net=a_net, f_u=f_u, gamma_m2=gamma_m2)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"N_{u,Rd} = 0.9 \cdot \frac{A_{net} \cdot f_u}{\gamma_{M2}} = "
                r"0.9 \cdot \frac{2000.000 \cdot 400.000}{1.100} = 654545.455 \ N",
            ),
            ("short", r"N_{u,Rd} = 654545.455 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_net = 2000.0
        f_u = 400.0
        gamma_m2 = 1.1

        # Object to test
        latex = Form6Dot7DesignUltimateResistanceNetCrossSection(a_net=a_net, f_u=f_u, gamma_m2=gamma_m2).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 9.13 from EN 1992-1-1:2004: Chapter 9 - Detailling and specific rules."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, MM
from blueprints.validations import raise_if_negative


class Form9Dot13TensileForceToBeAnchored(Formula):
    """Class representing the formula 9.13 for the calculation of the tensile force to be anchored for."""

    label = "9.13"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        r: MM,
        z_e: MM,
        z_i: MM,
    ) -> None:
        r"""[$F_s$] Tensile force to be anchored [$kN$].

        EN 1992-1-1:2004 art.9.8.2.2(2) - Formula (9.13)

        Parameters
        ----------
        r: MM
            [$R$] The resultant of ground pressure within x from figure 9.13 [$mm$].
        z_e: MM
            [$z_e$] The external lever arm, see figure 9.13, i.e. distance between the reinforcement and the horizontal force $F_c$ [$mm$].
        z_i: MM
            [$z_i$] Internal lever arm, see figure 9.13, i.e. distance between $R$ and the vertical force $N_{Ed}$ [$mm$].
        """
        super().__init__()
        self.r = r
        self.z_e = z_e
        self.z_i = z_i

    @staticmethod
    def _evaluate(
        r: MM,
        z_e: MM,
        z_i: MM,
    ) -> KN:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(r=r, z_e=z_e, z_i=z_i)
        return r * z_e / z_i

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 9.13."""
        return LatexFormula(
            return_symbol=r"F_s",
            result=f"{self:.{n}f}",
            equation=r"R \cdot z_e / z_i",
            numeric_equation=rf"{self.r:.{n}f} \cdot {self.z_e:.{n}f} / {self.z_i:.{n}f}",
            comparison_operator_label="=",
        )



"""Testing formula 6.60 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_60 import (
    Form6Dot60DesignValueCompressiveStressResistance,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot60DesignValueCompressiveStressResistance:
    """Validation for formula 6.60 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        k_1 = 0.8
        nu_prime = 0.6
        f_cd = 30.0

        # Object to test
        formula = Form6Dot60DesignValueCompressiveStressResistance(k_1=k_1, nu_prime=nu_prime, f_cd=f_cd)

        # Expected result, manually calculated
        manually_calculated_result = 14.4  # MPa

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("k_1", "nu_prime", "f_cd"),
        [
            (-0.8, 0.6, 30.0),  # k_1 is negative
            (0.8, -0.6, 30.0),  # nu_prime is negative
            (0.8, 0.6, -30.0),  # f_cd is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, k_1: float, nu_prime: float, f_cd: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot60DesignValueCompressiveStressResistance(k_1=k_1, nu_prime=nu_prime, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\sigma_{Rd,max} = k_1 \cdot \nu' \cdot f_{cd} = 0.800 \cdot 0.600 \cdot 30.000 = 14.400 \ MPa",
            ),
            ("short", r"\sigma_{Rd,max} = 14.400 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        k_1 = 0.8
        nu_prime = 0.6
        f_cd = 30.0

        # Object to test
        latex = Form6Dot60DesignValueCompressiveStressResistance(k_1=k_1, nu_prime=nu_prime, f_cd=f_cd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.41 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_41 import Form6Dot41BiaxialBendingCheck
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot41BiaxialBendingCheck:
    """Validation for formula 6.41 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        my_ed = 100.0
        m_n_y_rd = 200.0
        mz_ed = 50.0
        m_n_z_rd = 150.0
        alpha = 1.0
        beta = 1.0

        # Object to test
        formula = Form6Dot41BiaxialBendingCheck(
            my_ed=my_ed,
            m_n_y_rd=m_n_y_rd,
            mz_ed=mz_ed,
            m_n_z_rd=m_n_z_rd,
            alpha=alpha,
            beta=beta,
        )

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    def test_evaluation_not_ok(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        my_ed = 100.0
        m_n_y_rd = 200.0
        mz_ed = 50.0
        m_n_z_rd = 150.0
        alpha = 1.0
        beta = 0.63

        # Object to test
        formula = Form6Dot41BiaxialBendingCheck(
            my_ed=my_ed,
            m_n_y_rd=m_n_y_rd,
            mz_ed=mz_ed,
            m_n_z_rd=m_n_z_rd,
            alpha=alpha,
            beta=beta,
        )

        # Expected result, manually calculated
        expected_result = False

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("my_ed", "m_n_y_rd", "mz_ed", "m_n_z_rd", "alpha", "beta"),
        [
            (-100.0, 200.0, 50.0, 150.0, 1.0, 1.0),  # my_ed is negative
            (100.0, 200.0, -50.0, 150.0, 1.0, 1.0),  # mz_ed is negative
            (100.0, 200.0, 50.0, 150.0, -1.0, 1.0),  # alpha is negative
            (100.0, 200.0, 50.0, 150.0, 1.0, -1.0),  # beta is negative
        ],
    )
    def test_raise_error_when_negative_values_are_given(
        self, my_ed: float, m_n_y_rd: float, mz_ed: float, m_n_z_rd: float, alpha: float, beta: float
    ) -> None:
        """Test invalid negative values."""
        with pytest.raises(NegativeValueError):
            Form6Dot41BiaxialBendingCheck(
                my_ed=my_ed,
                m_n_y_rd=m_n_y_rd,
                mz_ed=mz_ed,
                m_n_z_rd=m_n_z_rd,
                alpha=alpha,
                beta=beta,
            )

    @pytest.mark.parametrize(
        ("my_ed", "m_n_y_rd", "mz_ed", "m_n_z_rd", "alpha", "beta"),
        [
            (100.0, 0.0, 50.0, 150.0, 1.0, 1.0),  # m_n_y_rd is zero
            (100.0, -200.0, 50.0, 150.0, 1.0, 1.0),  # m_n_y_rd is negative
            (100.0, 200.0, 50.0, 0.0, 1.0, 1.0),  # m_n_z_rd is zero
            (100.0, 200.0, 50.0, -150.0, 1.0, 1.0),  # m_n_z_rd is negative
        ],
    )
    def test_raise_error_when_invalid_denominator_values_are_given(
        self, my_ed: float, m_n_y_rd: float, mz_ed: float, m_n_z_rd: float, alpha: float, beta: float
    ) -> None:
        """Test invalid denominator values."""
        with pytest.raises(LessOrEqualToZeroError):
            Form6Dot41BiaxialBendingCheck(
                my_ed=my_ed,
                m_n_y_rd=m_n_y_rd,
                mz_ed=mz_ed,
                m_n_z_rd=m_n_z_rd,
                alpha=alpha,
                beta=beta,
            )

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left[ \frac{M_{y,Ed}}{M_{N,y,Rd}} \right]^{\alpha} + "
                r"\left[ \frac{M_{z,Ed}}{M_{N,z,Rd}} \right]^{\beta} \leq 1 \to "
                r"\left[ \frac{100.000}{200.000} \right]^{1.000} + "
                r"\left[ \frac{50.000}{150.000} \right]^{1.000} \leq 1 \to OK",
            ),
            (
                "complete_with_units",
                r"CHECK \to \left[ \frac{M_{y,Ed}}{M_{N,y,Rd}} \right]^{\alpha} + "
                r"\left[ \frac{M_{z,Ed}}{M_{N,z,Rd}} \right]^{\beta} \leq 1 \to "
                r"\left[ \frac{100.000 \ Nmm}{200.000 \ Nmm} \right]^{1.000} + "
                r"\left[ \frac{50.000 \ Nmm}{150.000 \ Nmm} \right]^{1.000} \leq 1 \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        my_ed = 100.0
        m_n_y_rd = 200.0
        mz_ed = 50.0
        m_n_z_rd = 150.0
        alpha = 1.0
        beta = 1.0

        # Object to test
        latex = Form6Dot41BiaxialBendingCheck(
            my_ed=my_ed,
            m_n_y_rd=m_n_y_rd,
            mz_ed=mz_ed,
            m_n_z_rd=m_n_z_rd,
            alpha=alpha,
            beta=beta,
        ).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 5.10 from EN 1993-5:2007 Chapter 5 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1993_5_2007 import EN_1993_5_2007
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot10ReductionFactorShearArea(Formula):
    """Class representing formula 5.10 for reduction factor for shear area."""

    label = "5.10"
    source_document = EN_1993_5_2007

    def __init__(
        self,
        v_ed: KN,
        v_pl_rd: KN,
    ) -> None:
        r"""[$\rho$] Calculate the reduction factor for shear area of the cross-section [$-$].

        EN 1993-5:2007(E) art.5.2.2(9) - Formula (5.10)

        Parameters
        ----------
        v_ed : KN
            [$V_{Ed}$] Design shear force in [$kN$].
        v_pl_rd : KN
            [$V_{pl,Rd}$] Plastic shear resistance in [$kN$].
        """
        super().__init__()
        self.v_ed: KN = v_ed
        self.v_pl_rd: KN = v_pl_rd

    @staticmethod
    def _evaluate(
        v_ed: KN,
        v_pl_rd: KN,
    ) -> DIMENSIONLESS:
        """Evaluates the formula for reduction factor for shear area."""
        raise_if_less_or_equal_to_zero(v_ed=v_ed, v_pl_rd=v_pl_rd)
        return (2 * v_ed / v_pl_rd - 1) ** 2

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 5.10."""
        return LatexFormula(
            return_symbol=r"\rho",
            result=f"{self:.{n}f}",
            equation=r"\left(2 \cdot \frac{V_{Ed}}{V_{pl,Rd}} - 1\right)^2",
            numeric_equation=rf"\left(2 \cdot \frac{{{self.v_ed}}}{{{self.v_pl_rd}}} - 1\right)^2",
            comparison_operator_label="=",
        )



"""Test Line class."""

from typing import Literal

import pytest
from shapely import Point

from blueprints.geometry.line import Line


class TestLine:
    """Test Line."""

    @pytest.fixture(autouse=True)
    def line(self) -> Line:
        """Fixture."""
        start_point = Point(0, 0, 0)
        end_point = Point(3, 4, 5)
        return Line(start_point=start_point, end_point=end_point)

    def test_error_same_points(self) -> None:
        """Test the error when the same points are given."""
        with pytest.raises(ValueError):
            Line(Point(0, 0, 0), Point(0, 0, 0))

    def test_initiate_with_no_z_value(self) -> None:
        """Test the initiation with no z value."""
        line = Line(Point(0, 0), Point(3, 4))
        assert line.start_point.has_z
        assert line.end_point.has_z

    def test_midpoint(self, line: Line) -> None:
        """Test the midpoint."""
        midpoint = line.midpoint
        assert midpoint == Point(1.5, 2.0, 2.5)

    def test_delta_x(self, line: Line) -> None:
        """Test the delta x."""
        assert line.delta_x == 3.0

    def test_delta_y(self, line: Line) -> None:
        """Test the delta y."""
        assert line.delta_y == 4.0

    def test_delta_z(self, line: Line) -> None:
        """Test the delta z."""
        assert line.delta_z == 5.0

    def test_length(self, line: Line) -> None:
        """Test the length."""
        assert line.length == pytest.approx(expected=7.07106, rel=1e-5)

    def test_angle(self, line: Line) -> None:
        """Test the angle."""
        angle = line.angle()
        assert angle == pytest.approx(expected=53.130, rel=1e-5)

    @pytest.mark.parametrize(
        ("distance", "reference", "x", "y", "z"),
        [
            (0.33, "start", 0.14, 0.18667, 0.2333),
            (0.33, "end", 2.86, 3.81333, 4.7667),
        ],
    )
    def test_get_internal_point(self, line: Line, distance: float, reference: Literal["start", "end"], x: float, y: float, z: float) -> None:
        """Test the internal points."""
        point = line.get_internal_point(distance=distance, reference=reference)
        assert (point.x, point.y, point.z) == pytest.approx(expected=(x, y, z), rel=1e-3)

    def test_get_internal_point_error(self, line: Line) -> None:
        """Test the internal point error."""
        with pytest.raises(ValueError):
            line.get_internal_point(distance=1.5, reference="middle")  # type: ignore[arg-type]
        with pytest.raises(ValueError):
            line.get_internal_point(distance=-1.5)
        with pytest.raises(ValueError):
            line.get_internal_point(distance=100_000)

    @pytest.mark.parametrize(
        ("distance", "direction", "expected_line"),
        [
            (-3.5355, "start", Line(Point(1.5, 2.0, 2.5), Point(3, 4, 5))),
            (-3.5355, "end", Line(Point(0, 0, 0), Point(1.5, 2.0, 2.5))),
            (1.0, "start", Line(Point(-0.42426, -0.565685, -0.707106), Point(3, 4, 5))),
            (1.0, "end", Line(Point(0, 0, 0), Point(3.424264, 4.565685, 5.707106))),
        ],
    )
    def test_adjust_length(self, line: Line, distance: float, direction: Literal["start", "end"], expected_line: Line) -> None:
        """Test the adjustment of the length."""
        adjusted_line = line.adjust_length(distance=distance, direction=direction)
        assert adjusted_line == expected_line

    def test_adjust_length_error(self, line: Line) -> None:
        """Test the adjustment of the length error."""
        with pytest.raises(ValueError):
            line.adjust_length(distance=1.5, direction="middle")  # type: ignore[arg-type]
        with pytest.raises(ValueError):
            line.adjust_length(distance=-100)

    def test_evenly_spaced_points(self, line: Line) -> None:
        """Test the evenly spaced points."""
        points = line.get_evenly_spaced_points(n=5)
        assert len(points) == 5
        assert points[0] == Point(0, 0, 0)
        assert points[-1] == Point(3, 4, 5)

    def test_evenly_spaced_points_error(self, line: Line) -> None:
        """Test the evenly spaced points error."""
        with pytest.raises(ValueError):
            line.get_evenly_spaced_points(n=1)
        with pytest.raises(TypeError):
            line.get_evenly_spaced_points(n=1.5)  # type: ignore[arg-type]

    def test_divide_into_n_lines(self, line: Line) -> None:
        """Test the division into n lines."""
        lines = line.divide_into_n_lines(n=5)
        assert len(lines) == 5
        assert lines[0].start_point == Point(0, 0, 0)
        assert lines[-1].end_point == Point(3, 4, 5)

    def test_divide_into_n_lines_error(self, line: Line) -> None:
        """Test the division into n lines error."""
        with pytest.raises(ValueError):
            line.divide_into_n_lines(n=1)
        with pytest.raises(TypeError):
            line.divide_into_n_lines(n=1.5)  # type: ignore[arg-type]

    def test_eq_true(self, line: Line) -> None:
        """Test the equality."""
        assert line == Line(Point(0, 0, 0), Point(3, 4, 5))

    def test_eq_false(self, line: Line) -> None:
        """Test the equality."""
        assert line != Line(Point(0, 0, 0), Point(3, 4, 6))

    def test_eq_not_line(self, line: Line) -> None:
        """Test the equality."""
        with pytest.raises(NotImplementedError):
            line == "Line"

    def test_repr(self, line: Line) -> None:
        """Test the representation."""
        assert repr(line) == "Line(POINT Z (0 0 0), POINT Z (3 4 5))"



"""Formula 6.4 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MM3, MM4, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot4ShearResistance(Formula):
    r"""Class representing formula 6.4 for the calculation of the shear resistance in regions uncracked in bending, [$V_{Rd,c}$]."""

    label = "6.4"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        i: MM4,
        b_w: MM,
        s: MM3,
        f_ctd: MPA,
        alpha_l: DIMENSIONLESS,
        sigma_cp: MPA,
    ) -> None:
        r"""[$V_{Rd,c}$] Shear resistance in regions uncracked in bending [$kN$].

        EN 1992-1-1:2004 art.6.2.2(2) - Formula (6.4)

        Parameters
        ----------
        i : MM4
            [$I$] Second moment of area [$mm^4$].
        b_w : MM
            [$b_w$] Width of the cross-section at the centroidal axis, see equation 6.16 and 6.17 [$mm$].
        s : MM3
            [$S$] First moment of area above and about the centroidal axis [$mm^3$].
        f_ctd : MPA
            [$f_{ctd}$] Design tensile strength of concrete [$MPa$].
        alpha_l : DIMENSIONLESS
            [$\alpha_l$] [$l_x / l_{pt2} \leq 1.0$] for pretensioned tendons, [$1.0$] for other types of prestressing [$-$].
        sigma_cp : MPA
            [$\sigma_{cp}$] Concrete compressive stress at the centroidal axis due to axial loading and/or prestressing [$MPa$].
        """
        super().__init__()
        self.i = i
        self.b_w = b_w
        self.s = s
        self.f_ctd = f_ctd
        self.alpha_l = alpha_l
        self.sigma_cp = sigma_cp

    @staticmethod
    def _evaluate(
        i: MM4,
        b_w: MM,
        s: MM3,
        f_ctd: MPA,
        alpha_l: DIMENSIONLESS,
        sigma_cp: MPA,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            i=i,
            b_w=b_w,
            s=s,
            f_ctd=f_ctd,
            alpha_l=alpha_l,
            sigma_cp=sigma_cp,
        )
        raise_if_less_or_equal_to_zero(s=s)

        return i * b_w / s * (f_ctd**2 + alpha_l * sigma_cp * f_ctd) ** 0.5

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.4."""
        return LatexFormula(
            return_symbol=r"V_{Rd,c}",
            result=f"{self:.{n}f}",
            equation=r"\frac{I \cdot b_w}{S} \cdot \sqrt{(f_{ctd})^2 + \alpha_l \cdot \sigma_{cp} \cdot f_{ctd}}",
            numeric_equation=rf"\frac{{{self.i:.{n}f} \cdot {self.b_w:.{n}f}}}{{{self.s:.{n}f}}} \cdot "
            rf"\sqrt{{({self.f_ctd:.{n}f})^2 + {self.alpha_l:.{n}f} \cdot {self.sigma_cp:.{n}f} \cdot {self.f_ctd:.{n}f}}}",
            comparison_operator_label="=",
            unit="N",
        )



"""Test the rebar module."""

import pytest

from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial
from blueprints.structural_sections.concrete.rebar import Rebar


class TestRebar:
    """Tests for the Rebar class."""

    @pytest.fixture
    def rebar(self) -> Rebar:
        """Return a Rebar instance."""
        return Rebar(
            diameter=20.0,
            x=0.0,
            y=0.0,
            material=ReinforcementSteelMaterial(),
            relative_start_position=0.1,
            relative_end_position=0.9,
        )

    def test_diameter(self, rebar: Rebar) -> None:
        """Test the diameter property of the Rebar class."""
        assert rebar.diameter == 20.0

    @pytest.mark.parametrize(
        "diameter",
        [
            0.0,
            -10.0,
        ],
    )
    def test_diameter_error(self, diameter: float) -> None:
        """Test the diameter property of the Rebar class."""
        with pytest.raises(ValueError):
            Rebar(
                diameter=diameter,
                x=0.0,
                y=0.0,
                material=ReinforcementSteelMaterial(),
            )

    def test_weight_per_meter(self, rebar: Rebar) -> None:
        """Test the weight_per_meter property of the Rebar class."""
        assert rebar.weight_per_meter == pytest.approx(expected=2.4661, rel=1e-2)

    def test_relative_start_position(self, rebar: Rebar) -> None:
        """Test the relative_start_position property of the Rebar class."""
        assert rebar.relative_start_position == 0.1

    def test_relative_end_position(self, rebar: Rebar) -> None:
        """Test the relative_end_position property of the Rebar class."""
        assert rebar.relative_end_position == 0.9

    @pytest.mark.parametrize(
        "relative_start_position",
        [
            1.1,
            -0.1,
        ],
    )
    def test_relative_start_position_error(self, relative_start_position: float) -> None:
        """Test the relative_start_position property of the Rebar class."""
        with pytest.raises(ValueError):
            Rebar(
                diameter=20.0,
                x=0.0,
                y=0.0,
                material=ReinforcementSteelMaterial(),
                relative_start_position=relative_start_position,
            )

    @pytest.mark.parametrize(
        "relative_end_position",
        [
            1.1,
            -0.1,
        ],
    )
    def test_relative_end_position_error(self, relative_end_position: float) -> None:
        """Test the relative_end_position property of the Rebar class."""
        with pytest.raises(ValueError):
            Rebar(
                diameter=20.0,
                x=0.0,
                y=0.0,
                material=ReinforcementSteelMaterial(),
                relative_end_position=relative_end_position,
            )



"""Formula 5.45 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN, ONE_OVER_M, M
from blueprints.validations import raise_if_negative


class Form5Dot45LossesDueToFriction(Formula):
    r"""Class representing formula 5.45 for the calculation of losses due to friction in post-tensioned tendons, [$\Delta P_{\mu}(x)$]."""

    label = "5.45"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        p_max: KN,
        mu: DIMENSIONLESS,
        theta: DIMENSIONLESS,
        k: ONE_OVER_M,
        x: M,
    ) -> None:
        r"""[$\Delta P_{\mu}(x)$] Losses due to friction [$kN$].

        EN 1992-1-1:2004 art.5.10.5.2(1) - Formula (5.45)

        Parameters
        ----------
        p_max : KN
            [$P_{max}$] Force at the active end during tensioning [$kN$].
        mu : DIMENSIONLESS
            [$\mu$] Coefficient of friction between the tendon and its duct [$-$].
        theta : DIMENSIONLESS
            [$\theta$] Sum of the angular displacements over a distance x (irrespective of direction or sign) [$-$].
        k : ONE_OVER_M
            [$k$] Unintentional angular displacement for internal tendons (per unit length) [$1/m$].
        x : M
            [$x$] Distance along tendon from point where the prestressing force equals Pmax (force at active end during tensioning) [$m$].
        """
        super().__init__()
        self.p_max = p_max
        self.mu = mu
        self.theta = theta
        self.k = k
        self.x = x

    @staticmethod
    def _evaluate(
        p_max: KN,
        mu: DIMENSIONLESS,
        theta: DIMENSIONLESS,
        k: ONE_OVER_M,
        x: M,
    ) -> KN:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            p_max=p_max,
            mu=mu,
            theta=theta,
            k=k,
            x=x,
        )

        return p_max * (1 - (np.e ** (-mu * (theta + k * x))))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.45."""
        return LatexFormula(
            return_symbol=r"\Delta P_{\mu}(x)",
            result=f"{self:.{n}f}",
            equation=r"P_{max} \cdot \left( 1 - e^{-\mu \cdot (\theta + k \cdot x)} \right)",
            numeric_equation=rf"{self.p_max:.{n}f} \cdot \left( 1 - e^{{-{self.mu:.{n}f} \cdot ({self.theta:.{n}f} + "
            rf"{self.k:.{n}f} \cdot {self.x:.{n}f})}} \right)",
            comparison_operator_label="=",
            unit="kN",
        )



"""Formula 7.7 from EN 1995-1-1:2004."""

from blueprints.codes.eurocode.en_1995_1_1_2004 import EN_1995_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, HZ, NM2_M, M
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form7Dot7NumberOfFOVibrations(Formula):
    r"""Class representing formula 7.7 for the calculations of the number of first-order vibrations with a natural frequency less than 40Hz."""

    label = "7.7"
    source_document = EN_1995_1_1_2004
    # frequency limit to check
    f_40 = 40

    def __init__(self, f_1: HZ, b: M, length: M, ei_l: NM2_M, ei_b: NM2_M) -> None:
        r"""[$n_{40}$] The number of first-order vibrations with a natural frequency less than 40Hz [$-$].

        EN 1995-1-1:2004 art 7.3.3(5) - Formula (7.7)

        Parameters
        ----------
        f_1 : HZ
            [$f_{1}$] Natural frequency of rectangular floor, laid freely on all four sides [$Hz$].
        length : M
            [$l$] Span of the floor [$m$].
        b : M
            [$b$] Width of the floor [$m$].
        ei_l : NM2_M
            [$(EI)_{l}$] Equivalent bending stiffness of the floor around the axis perpendicular to the longitudinal axis of the beam [$Nm^{2}/m$].
        ei_b : NM2_M
            [$(EI)_{l}$] Equivalent bending stiffness of the floor around the axis parallel to the longitudinal axis of the beam, with $EI_{b} < EI_{l}$ [$Nm^{2}/m$].

        Returns
        -------
        None
        """  # noqa: E501
        super().__init__()
        self.f_1 = f_1
        self.length = length
        self.b = b
        self.ei_l = ei_l
        self.ei_b = ei_b

    @staticmethod
    def _evaluate(f_1: HZ, b: M, length: M, ei_l: NM2_M, ei_b: NM2_M) -> DIMENSIONLESS:
        """Evaluate the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(f_1=f_1, b=b, length=length, ei_l=ei_l, ei_b=ei_b)
        if f_1 > Form7Dot7NumberOfFOVibrations.f_40:
            raise ValueError(f"Value of $f_1$ ({f_1:.3f}) exceeds the allowed limit of {Form7Dot7NumberOfFOVibrations.f_40}.")
        if not ei_b < ei_l:
            raise ValueError(rf"$EI_l$ ({ei_l:.2f}) must be bigger than $EI_b$ ({ei_b:.2f}).")
        return (((40 / f_1) ** 2 - 1) * (b / length) ** 4 * ei_l / ei_b) ** 0.25

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 7.7."""
        eq_for: str = (
            r"\left\{\left ( \left (\frac{40}{f_1} \right )^2 -1 \right)\left ( \frac{b}{l } \right )^4 \frac{(EI)_l}{(EI)_b}\right\}^{0.25}"
        )
        repl_symb = {
            "(EI)_l": f"{self.ei_l:.{n}f}",
            "(EI)_b": f"{self.ei_b:.{n}f}",
            "f_1": f"{self.f_1:.{n}f}",
            "l ": f"{self.length:.{n}f}",
            "b": f"{self.b:.{n}f}",
        }
        return LatexFormula(
            return_symbol=r"n_{40}",
            result=f"{self:.{n}f}",
            equation=eq_for,
            numeric_equation=latex_replace_symbols(eq_for, repl_symb),
        )



"""Formula 6.24 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot24DesignShearStress(Formula):
    r"""Class representing formula 6.24 for the calculation of the design shear stress in the interface, [$v_{Edi}$]."""

    label = "6.24"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        beta: DIMENSIONLESS,
        v_ed: N,
        z: MM,
        b_i: MM,
    ) -> None:
        r"""[$v_{Edi}$] Design shear stress in the interface [$MPa$].

        EN 1992-1-1:2004 art.6.2.5(1) - Formula (6.24)

        Parameters
        ----------
        beta : DIMENSIONLESS
            [$\beta$] Ratio of the longitudinal force in the new concrete area and the total longitudinal force either in
            the compression or tension zone, both calculated for the section considered [$-$].
        v_ed : N
            [$V_{Ed}$] Transverse shear force [$N$].
        z : MM
            [$z$] Lever arm of composite section [$mm$].
        b_i : MM
            [$b_{i}$] Width of the interface (see Figure 6.8) [$mm$].
        """
        super().__init__()
        self.beta = beta
        self.v_ed = v_ed
        self.z = z
        self.b_i = b_i

    @staticmethod
    def _evaluate(
        beta: DIMENSIONLESS,
        v_ed: N,
        z: MM,
        b_i: MM,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            beta=beta,
            v_ed=v_ed,
        )
        raise_if_less_or_equal_to_zero(z=z, b_i=b_i)

        return beta * v_ed / (z * b_i)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.24."""
        return LatexFormula(
            return_symbol=r"v_{Edi}",
            result=f"{self:.{n}f}",
            equation=r"\beta \cdot \frac{V_{Ed}}{z \cdot b_{i}}",
            numeric_equation=rf"{self.beta:.{n}f} \cdot \frac{{{self.v_ed:.{n}f}}}{{{self.z:.{n}f} \cdot {self.b_i:.{n}f}}}",
            comparison_operator_label="=",
            unit="MPa",
        )



"""Testing formula 7.14 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_14 import Form7Dot14MaximumCrackSpacing
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot14MaximumCrackSpacing:
    """Validation for formula 7.14 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        h = 500.0
        x = 200.0

        # Object to test
        formula = Form7Dot14MaximumCrackSpacing(h=h, x=x)

        # Expected result, manually calculated
        manually_calculated_result = 390.0  # mm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("h", "x"),
        [
            (-500.0, 200.0),  # h is negative
            (500.0, -200.0),  # x is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, h: float, x: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form7Dot14MaximumCrackSpacing(h=h, x=x)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"s_{r,max} = 1.3 \cdot (h - x) = 1.3 \cdot (500.000 - 200.000) = 390.000 \ mm",
            ),
            ("short", r"s_{r,max} = 390.000 \ mm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        h = 500.0
        x = 200.0

        # Object to test
        latex = Form7Dot14MaximumCrackSpacing(h=h, x=x).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Tests for Rectangular Reinforced Concrete Sections."""

from typing import Literal

import pytest
from matplotlib import pyplot as plt
from shapely import LineString, Polygon

from blueprints.structural_sections.concrete.rebar import Rebar
from blueprints.structural_sections.concrete.reinforced_concrete_sections.rectangular import RectangularReinforcedCrossSection
from blueprints.structural_sections.concrete.reinforced_concrete_sections.reinforcement_configurations import ReinforcementByQuantity
from blueprints.structural_sections.concrete.stirrups import StirrupConfiguration


class TestRectangularReinforcedCrossSection:
    """Tests for the RectangularReinforcedCrossSection class."""

    def test_add_stirrup_along_edges(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the add_stirrup_along_edges method."""
        stirrup = rectangular_reinforced_cross_section.add_stirrup_along_edges(
            diameter=8,
            distance=150,
            material=rectangular_reinforced_cross_section.get_present_steel_materials()[0],
        )
        assert stirrup in rectangular_reinforced_cross_section.stirrups

    def test_add_stirrup_in_center(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the add_stirrup_in_center method."""
        stirrup = rectangular_reinforced_cross_section.add_stirrup_in_center(
            width=200,
            diameter=8,
            distance=150,
            material=rectangular_reinforced_cross_section.get_present_steel_materials()[0],
        )
        assert stirrup in rectangular_reinforced_cross_section.stirrups

    @pytest.mark.parametrize(
        "edge",
        [
            "upper",
            "right",
            "lower",
            "left",
        ],
    )
    def test_add_longitudinal_reinforcement_by_quantity(
        self,
        rectangular_reinforced_cross_section: RectangularReinforcedCrossSection,
        edge: Literal["upper", "right", "lower", "left"],
    ) -> None:
        """Test the add_longitudinal_reinforcement_by_quantity method."""
        rectangular_reinforced_cross_section.add_longitudinal_reinforcement_by_quantity(
            n=5,
            diameter=14,
            material=rectangular_reinforced_cross_section.get_present_steel_materials()[0],
            edge=edge,
            cover=23,
            corner_offset=40,
        )
        assert len(rectangular_reinforced_cross_section.longitudinal_rebars) == 15

    def test_add_longitudinal_reinforcement_by_quantity_wrong_edge(
        self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection
    ) -> None:
        """Test the add_longitudinal_reinforcement_by_quantity method with wrong edge."""
        with pytest.raises(ValueError):
            rectangular_reinforced_cross_section.add_longitudinal_reinforcement_by_quantity(
                n=5,
                diameter=14,
                material=rectangular_reinforced_cross_section.get_present_steel_materials()[0],
                edge="wrong",  # type: ignore[arg-type]
            )

    def test_plot(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the plot method."""
        plot = rectangular_reinforced_cross_section.plot(show=False, center_line_style={"linewidth": 0.85})
        assert isinstance(plot, plt.Figure)

    def test_reinforcement_weight_longitudinal_bars(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the reinforcement_weight_longitudinal_bars method."""
        expected_weight = 46.38828588400884  # kg/m
        assert rectangular_reinforced_cross_section.reinforcement_weight_longitudinal_bars == pytest.approx(expected=expected_weight, rel=1e-2)

    def test_reinforcement_weight_stirrups(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the reinforcement_weight_stirrups method."""
        expected_weight = 18.1087  # kg/m
        assert rectangular_reinforced_cross_section.reinforcement_weight_stirrups == pytest.approx(expected=expected_weight, rel=1e-2)

    def test_reinforcement_weight(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the reinforcement_weight method."""
        expected_weight = 64.4969  # kg/m
        assert rectangular_reinforced_cross_section.reinforcement_weight == pytest.approx(expected=expected_weight, rel=1e-2)

    def test_reinforcement_area_longitudinal_bars(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the reinforcement_area_longitudinal_bars method."""
        expected_area = 5909.3357  # mm²/m
        assert rectangular_reinforced_cross_section.reinforcement_area_longitudinal_bars == pytest.approx(expected=expected_area, rel=1e-2)

    def test_concrete_volume(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the concrete_volume method."""
        expected_volume = 0.8  # m³/m
        assert rectangular_reinforced_cross_section.concrete_volume == pytest.approx(expected=expected_volume, rel=1e-2)

    def test_weight_per_volume(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the weight_per_volume method."""
        expected_weight_per_volume = 80.6211  # kg/m³
        assert rectangular_reinforced_cross_section.weight_per_volume == pytest.approx(expected=expected_weight_per_volume, rel=1e-2)

    def test_add_longitudinal_rebar_wrong_position(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the add_longitudinal_rebar method with wrong position."""
        with pytest.raises(ValueError):
            rectangular_reinforced_cross_section.add_longitudinal_rebar(
                rebar=Rebar(
                    diameter=12,
                    x=2500,
                    y=1000,
                    material=rectangular_reinforced_cross_section.get_present_steel_materials()[0],
                )
            )

    def test_add_stirrup_configuration_wrong_position(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the add_stirrup_configuration method with wrong position."""
        with pytest.raises(ValueError):
            rectangular_reinforced_cross_section.add_stirrup_configuration(
                stirrup=StirrupConfiguration(
                    geometry=Polygon([(0, 0), (0, 100), (5000, 5000), (100, 0)]),
                    diameter=8,
                    distance=150,
                    material=rectangular_reinforced_cross_section.get_present_steel_materials()[0],
                ),
            )

    def test_add_reinforcement_configuration_by_linestring(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the add_reinforcement_configuration method with a linestring."""
        linestring = LineString([(-300, 200), (300, 200)])
        rectangular_reinforced_cross_section.add_reinforcement_configuration(
            line=linestring,
            configuration=ReinforcementByQuantity(
                diameter=12,
                n=3,
                material=rectangular_reinforced_cross_section.get_present_steel_materials()[0],
            ),
        )
        assert len(rectangular_reinforced_cross_section.longitudinal_rebars) == 13

    def test_rebar_not_in_cross_section(self, rectangular_reinforced_cross_section: RectangularReinforcedCrossSection) -> None:
        """Test the add_longitudinal_rebar method with a rebar not in the cross-section."""
        rebar = Rebar(
            diameter=12,
            x=2500,
            y=1000,
            material=rectangular_reinforced_cross_section.get_present_steel_materials()[0],
        )
        rectangular_reinforced_cross_section._single_longitudinal_rebars.append(rebar)  # noqa: SLF001
        with pytest.raises(ValueError):
            _ = rectangular_reinforced_cross_section.longitudinal_rebars



"""Test for the UNP enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.unp import UNP


class TestUNP:
    """Tests for the UNP enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert UNP.UNP80.value == ("UNP80", 80, 45, 6, 8, 8, 4, 8)
        assert UNP.UNP400.value == ("UNP400", 400, 110, 14, 18, 18, 9, 5)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "UNP80" in [e.value[0] for e in UNP]
        assert "UNP400" in [e.value[0] for e in UNP]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in UNP]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = UNP.UNP80
        assert profile.top_flange_total_width == 45
        assert profile.top_flange_thickness == 8
        assert profile.bottom_flange_total_width == 45
        assert profile.bottom_flange_thickness == 8
        assert profile.total_height == 80
        assert profile.web_thickness == 6
        assert profile.root_fillet_radius == 8
        assert profile.toe_radius == 4
        assert profile.slope == 8



"""Module for the concrete exposure classes
according to Table 4.1 from EN 1992-1-1: Chapter 4 - Durability and cover to reinforcement.
"""

import re
from abc import abstractmethod
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from enum import Enum
from functools import total_ordering
from typing import ClassVar, Self

from blueprints.utils.abc_enum_meta import ABCEnumMeta


@total_ordering
class Exposure(Enum, metaclass=ABCEnumMeta):
    """Parent class for individual exposure classes.

    This class handles the ordering/comparison operations, that's why it is decorated with total_ordering (As recommended by PEP8).
    On top of that, it handles a couple of methods which will be used by its subclasses.
    """

    def __eq__(self, other: object) -> bool:
        """Definition of '==' operator for the comparison of the severity of the exposure classifications.

        Parameters
        ----------
        self : Exposure/ subclass of Exposure
            First argument for the comparison.
        other : object
            Second argument for the comparison.

        Raises
        ------
        TypeError
            If different types are being compared.

        Returns
        -------
        Boolean
            True if both arguments are of the same severity (In this case they will both be literally the same).
        """
        if not isinstance(other, self.__class__):
            raise TypeError("Only the same exposure class types can be compared with each other!")
        _self_severity = int(self.value[-1]) if self.value != "Not applicable" else 0
        _other_severity = int(other.value[-1]) if other.value != "Not applicable" else 0
        return _self_severity == _other_severity

    def __gt__(self, other: Self) -> bool:
        """Definition of '>' operator for the comparison of the severity of the exposure classifications.

        Parameters
        ----------
        self : Exposure/ subclass of Exposure
            First argument for the comparison.
        other : object
            Second argument for the comparison.

        Raises
        ------
        TypeError
            If different types are being compared.

        Returns
        -------
        Boolean
            True if the first argument is more severe than the second argument.
        """
        if isinstance(other, self.__class__):
            _self_severity = int(self.value[-1]) if self.value != "Not applicable" else 0
            _other_severity = int(other.value[-1]) if other.value != "Not applicable" else 0
            return _self_severity > _other_severity
        raise TypeError("Only the same exposure class types can be compared with each other!")

    @classmethod
    def options(cls) -> list[str]:
        """Return all the possible options within a subclass.

        Returns
        -------
        list[str]
            all the possible class designations within a specific exposure class
        """
        return [m._value_ for m in cls.__members__.values()]

    @staticmethod
    @abstractmethod
    def exposure_class_description() -> str:
        """Description of subclasses to be implemented in each subclass.

        Returns
        -------
        str
            description of the specific exposure class
        """

    @abstractmethod
    def description_of_the_environment(self) -> str:
        """Description of the environment based on the instance.

        Returns
        -------
        str
            description of the environment based on the instance
        """

    @classmethod
    def snake_case(cls) -> str:
        """Converts the name of a subclass to snake_case which can be used in a parametrization class.

        Returns
        -------
        str
            the name of a subclass in snake_case
        """
        return re.sub(r"(?<!^)(?=[A-Z])", "_", cls.__name__).lower()

    @classmethod
    def notation(cls) -> str:
        """Returns the notation of the exposure class.

        Returns
        -------
        str
            notation of the exposure class, e.g. "XC", "XD", etc.
        """
        options = cls.options()
        # Look for this pattern in the options: Uppercase letters followed by at least one digit
        # If the pattern is found, return the Uppercase letters part
        for option in options:
            match = re.match(r"([A-Z]+)(\d+)", option)
            if match:
                return match.group(1)
        raise ValueError(
            f"No valid notation found for {cls.__name__}.\n"
            "Either adhere to the pattern of Uppercase letters followed by at least one digit for Enum values, "
            "or implement the notation method in the subclass."
        )


@dataclass(frozen=True)
class ExposureClassesBase:
    """Parent class which serves as a container for the Exposure classes.

    Exposure classes related to environmental conditions in accordance with EN 206-1
    """

    _not_applicable_key: ClassVar[str] = "Not applicable"
    """Key for the 'Not applicable' exposure class. Can be overridden in subclasses if needed."""

    @property
    def no_risk(self) -> bool:
        """Check if all exposure classes are 'Not applicable' (specified by _not_applicable_key).

        This represents X0 class designation according to table 4.1 from EN 1992-1-1:2004.

        Returns
        -------
        bool
            True if all exposure classes are euqal to _not_applicable_key ("Not applicable" by default)
        """
        return all(exposure_class.value == self.__class__._not_applicable_key for exposure_class in self.__dict__.values())  # noqa: SLF001

    def __str__(self) -> str:
        """String representation of the ExposureClasses object.

        Returns
        -------
        str
            String representation of the ExposureClasses object
        """
        return "X0" if self.no_risk else ", ".join(enum.value for enum in self.__dict__.values() if enum.value != self.__class__._not_applicable_key)  # noqa: SLF001

    def __iter__(self) -> Iterator[Exposure]:
        """Iterator for the ExposureClasses object.

        Returns
        -------
        Iterator[Exposure]
            Iterator for the ExposureClasses object
        """
        return iter(self.__dict__.values())

    @classmethod
    def from_exposure_list(cls, exposure_classes: Sequence[str]) -> Self:
        """Create an instance from a sequence of exposure classes.

        Examples
        --------
        >>> exposure_classes = ["XC1", "XD1", "XS1"]
        >>> ConcreteExposureClasses().from_exposure_list(exposure_classes)
        ConcreteExposureClasses(
            carbonation=<Carbonation.XC1: 'XC1'>,
            chloride=<Chloride.XD1: 'XD1'>,
            chloride_seawater=<ChlorideSeawater.XS1: 'XS1'>,
            freeze_thaw=<FreezeThaw.NA: 'Not applicable'>,
            chemical=<Chemical.NA: 'Not applicable'>
        )

        Parameters
        ----------
        exposure_classes : Sequence[str]
            sequence of exposure classes, order is not important.
            If an exposure class is not provided, but is defined in the __init__, it is set to the value of _not_applicable_key ("Not applicable").
            You can use capital letters or lowercase letters, the method is case-insensitive.
            For example, "XC1" and "xc1" are both valid.

        Returns
        -------
        Self
            instance created from the list
        """
        exposures: dict[str, Exposure] = {}
        classifications: dict[str, type[Exposure]] = {arg.notation(): arg for kw, arg in cls.__init__.__annotations__.items() if kw != "return"}

        for exposure_str in exposure_classes:
            classification = classifications.get(exposure_str[:2].upper())
            if classification is None:
                raise ValueError(f"Invalid exposure class: '{exposure_str}'")
            classification_name = classification.snake_case()
            if classification_name in exposures:
                raise ValueError(f"Duplication Error: There are multiple instances of '{classification.__name__}' class.")
            exposures[classification_name] = classification[exposure_str.upper()]

        for classification in classifications.values():
            classification_name = classification.snake_case()
            if classification_name not in exposures:
                exposures.setdefault(classification_name, classification(cls._not_applicable_key))

        return cls(**exposures)



"""Formulas 6.37 and 6.38 from EN 1993-1-1:2005: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM, MM2, NMM, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot37Dot38MomentReduction(Formula):
    r"""Class representing formulas 6.37 and 6.38 for the calculation of reduced bending moment."""

    label = "6.37/6.38"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        mpl_z_rd: NMM,
        a: DIMENSIONLESS,
        n: DIMENSIONLESS,
    ) -> None:
        r"""[$M_{N,z,Rd}$] Reduced bending moment [$Nmm$].

        EN 1993-1-1:2005 art.6.2.9.1(5) - Formulas (6.37 and 6.38)

        Parameters
        ----------
        mpl_z_rd : NMM
            [$M_{pl,z,Rd}$] Plastic bending moment about the z-axis [$Nmm$].
        a : DIMENSIONLESS
            Reduction factor for cross-sectional area, see equation 6.38a (`Form6Dot38A`) [-].
        n : DIMENSIONLESS
            Axial force ratio, see equation 6.38n (`Form6Dot38N`) [-].
        """
        super().__init__()
        self.mpl_z_rd = mpl_z_rd
        self.a = a
        self.n = n

    @staticmethod
    def _evaluate(
        mpl_z_rd: NMM,
        a: DIMENSIONLESS,
        n: DIMENSIONLESS,
    ) -> NMM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(mpl_z_rd=mpl_z_rd, n=n)
        raise_if_less_or_equal_to_zero(a=a)

        denominator = 1 - a
        raise_if_less_or_equal_to_zero(denominator=denominator)

        if n <= a:
            return mpl_z_rd
        return mpl_z_rd * (1 - ((n - a) / denominator) ** 2)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formulas 6.37 and 6.38."""
        _equation: str = (
            r"\begin{cases} M_{pl,z,Rd} & \text{if } n \leq a \\ "
            r"M_{pl,z,Rd} \cdot \left[1 - \left(\frac{ n - a}{1 - a}\right)^2\right] & \text{if } n > a \end{cases}"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "M_{pl,z,Rd}": f"{self.mpl_z_rd:.{n}f}",
                " n": f" {self.n:.{n}f}",
                " a": f" {self.a:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                "M_{pl,z,Rd}": rf"{self.mpl_z_rd:.{n}f} \ Nmm",
                " n": rf" {self.n:.{n}f}",
                " a": rf" {self.a:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"M_{N,z,Rd}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="=",
            unit="Nmm",
        )


class Form6Dot38N(Formula):
    r"""Class representing formula 6.38 for the calculation of [$n$]."""

    label = "6.38n"
    source_document = EN_1993_1_1_2005

    def __init__(self, n_ed: N, n_pl_rd: N) -> None:
        r"""[$n$] Axial force ratio [dimensionless].

        EN 1993-1-1:2005 art.6.2.9(5) - Formula (6.38n)

        Parameters
        ----------
        n_ed : N
            [$N_{Ed}$] Design axial force [$N$].
        n_pl_rd : N
            [$N_{pl,Rd}$] Plastic resistance of the cross-section [$N$].
        """
        super().__init__()
        self.n_ed = n_ed
        self.n_pl_rd = n_pl_rd

    @staticmethod
    def _evaluate(n_ed: N, n_pl_rd: N) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(n_pl_rd=n_pl_rd)
        raise_if_negative(n_ed=n_ed)
        return n_ed / n_pl_rd

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.38n."""
        _equation: str = r"\frac{N_{Ed}}{N_{pl,Rd}}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"N_{Ed}": f"{self.n_ed:.{n}f}",
                r"N_{pl,Rd}": f"{self.n_pl_rd:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                r"N_{Ed}": rf"{self.n_ed:.{n}f} \ N",
                r"N_{pl,Rd}": rf"{self.n_pl_rd:.{n}f} \ N",
            },
            True,
        )
        return LatexFormula(
            return_symbol=r"n",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="=",
            unit="-",
        )


class Form6Dot38A(Formula):
    r"""Class representing formula 6.38 for the calculation of [$a$]."""

    label = "6.38a"
    source_document = EN_1993_1_1_2005

    def __init__(self, capital_a: MM2, b: MM, tf: MM) -> None:
        r"""[$a$] Reduction factor for cross-sectional area [dimensionless].

        EN 1993-1-1:2005 art.6.2.9(5) - Formula (6.38a)

        Parameters
        ----------
        capital_a : MM2
            [$A$] Cross-sectional area [$mm^2$].
        b : MM
            [$b$] Width of the cross-section [$mm$].
        tf : MM
            [$t_f$] Thickness of the flange [$mm$].
        """
        super().__init__()
        self.capital_a = capital_a
        self.b = b
        self.tf = tf

    @staticmethod
    def _evaluate(capital_a: MM2, b: MM, tf: MM) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(capital_a=capital_a)
        raise_if_negative(b=b, tf=tf)
        a = (capital_a - 2 * b * tf) / capital_a
        return min(a, 0.5)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.38a."""
        _equation: str = r"\min\left(\frac{A - 2 \cdot b \cdot t_f}{A}, 0.5\right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A": f"{self.capital_a:.{n}f}",
                r"b": f"{self.b:.{n}f}",
                r"t_f": f"{self.tf:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                r"A": rf"{self.capital_a:.{n}f} \ mm^2",
                r"b": rf"{self.b:.{n}f} \ mm",
                r"t_f": rf"{self.tf:.{n}f} \ mm",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"a",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="=",
            unit="-",
        )



"""Testing formula 9.14 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_9_detailling_and_specific_rules.formula_9_14 import (
    Form9Dot14SplittingForceColumnOnRock,
)
from blueprints.validations import NegativeValueError


class TestForm9Dot14SplittingForceColumnOnRock:
    """Validation for formula 9.14 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c = 50  # mm
        h = 100  # mm
        n_ed = 200  # kN
        form_9_14 = Form9Dot14SplittingForceColumnOnRock(c=c, h=h, n_ed=n_ed)

        # Expected result, manually calculated
        manually_calculated_result = 25

        assert form_9_14 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_c_is_given(self) -> None:
        """Test the evaluation of the result."""
        c = -50  # mm
        h = 100  # mm
        n_ed = 200  # kN

        with pytest.raises(NegativeValueError):
            Form9Dot14SplittingForceColumnOnRock(c=c, h=h, n_ed=n_ed)

    def test_raise_error_when_negative_h_is_given(self) -> None:
        """Test the evaluation of the result."""
        c = 50  # mm
        h = -100  # mm
        n_ed = 200  # kN

        with pytest.raises(NegativeValueError):
            Form9Dot14SplittingForceColumnOnRock(c=c, h=h, n_ed=n_ed)

    def test_raise_error_when_negative_n_ed_is_given(self) -> None:
        """Test the evaluation of the result."""
        c = 50  # mm
        h = 100  # mm
        n_ed = -200  # kN

        with pytest.raises(NegativeValueError):
            Form9Dot14SplittingForceColumnOnRock(c=c, h=h, n_ed=n_ed)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"F_s = 0.25 \cdot ( 1 - c / h ) \cdot N_{Ed} = 0.25 \cdot ( 1 - 50.00 / 100.00 ) \cdot 200.00 = 25.00",
            ),
            ("short", r"F_s = 25.00"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        c = 50  # mm
        h = 100  # mm
        n_ed = 200  # kN

        # Object to test
        form_9_14_latex = Form9Dot14SplittingForceColumnOnRock(c=c, h=h, n_ed=n_ed).latex()

        actual = {"complete": form_9_14_latex.complete, "short": form_9_14_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Test the RHSCF enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.rhscf import RHSCF


class TestRHSCF:
    """Tests for the RHSCF enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert RHSCF.RHSCF40x20_2.value == ("RHSCF40x20x2", 40, 20, 2, 4, 2)
        assert RHSCF.RHSCF400x300_16.value == ("RHSCF400x300x16", 400, 300, 16, 48, 32)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "RHSCF40x20x2" in [e.value[0] for e in RHSCF]
        assert "RHSCF400x300x16" in [e.value[0] for e in RHSCF]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in RHSCF]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = RHSCF.RHSCF40x20_2
        assert profile.alias == "RHSCF40x20x2"
        assert profile.total_height == 40
        assert profile.total_width == 20
        assert profile.thickness == 2
        assert profile.outer_radius == 4
        assert profile.inner_radius == 2



"""Testing formula 5.21 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_21 import Form5Dot21NominalStiffness
from blueprints.validations import NegativeValueError


class TestForm5Dot21NominalStiffness:
    """Validation for formula 5.21 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        k_c = 0.8  # -
        e_cd = 30000  # MPa
        i_c = 5000  # mm^4
        k_s = 0.2  # -
        e_s = 200000  # MPa
        i_s = 1000  # mm^4

        # Object to test
        form_5_21 = Form5Dot21NominalStiffness(k_c=k_c, e_cd=e_cd, i_c=i_c, k_s=k_s, e_s=e_s, i_s=i_s)

        # Expected result, manually calculated
        manually_calculated_result = 160000000  # Nmm^2

        assert form_5_21 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_k_c_is_negative(self) -> None:
        """Test negative value for k_c."""
        k_c = -0.8
        e_cd = 30000
        i_c = 5000
        k_s = 0.2
        e_s = 200000
        i_s = 1000

        with pytest.raises(NegativeValueError):
            Form5Dot21NominalStiffness(k_c=k_c, e_cd=e_cd, i_c=i_c, k_s=k_s, e_s=e_s, i_s=i_s)

    def test_raise_error_when_e_cd_is_negative(self) -> None:
        """Test negative value for e_cd."""
        k_c = 0.8
        e_cd = -30000
        i_c = 5000
        k_s = 0.2
        e_s = 200000
        i_s = 1000

        with pytest.raises(NegativeValueError):
            Form5Dot21NominalStiffness(k_c=k_c, e_cd=e_cd, i_c=i_c, k_s=k_s, e_s=e_s, i_s=i_s)

    def test_raise_error_when_i_c_is_negative(self) -> None:
        """Test negative value for i_c."""
        k_c = 0.8
        e_cd = 30000
        i_c = -5000
        k_s = 0.2
        e_s = 200000
        i_s = 1000

        with pytest.raises(NegativeValueError):
            Form5Dot21NominalStiffness(k_c=k_c, e_cd=e_cd, i_c=i_c, k_s=k_s, e_s=e_s, i_s=i_s)

    def test_raise_error_when_k_s_is_negative(self) -> None:
        """Test negative value for k_s."""
        k_c = 0.8
        e_cd = 30000
        i_c = 5000
        k_s = -0.2
        e_s = 200000
        i_s = 1000

        with pytest.raises(NegativeValueError):
            Form5Dot21NominalStiffness(k_c=k_c, e_cd=e_cd, i_c=i_c, k_s=k_s, e_s=e_s, i_s=i_s)

    def test_raise_error_when_e_s_is_negative(self) -> None:
        """Test negative value for e_s."""
        k_c = 0.8
        e_cd = 30000
        i_c = 5000
        k_s = 0.2
        e_s = -200000
        i_s = 1000

        with pytest.raises(NegativeValueError):
            Form5Dot21NominalStiffness(k_c=k_c, e_cd=e_cd, i_c=i_c, k_s=k_s, e_s=e_s, i_s=i_s)

    def test_raise_error_when_i_s_is_negative(self) -> None:
        """Test negative value for i_s."""
        k_c = 0.8
        e_cd = 30000
        i_c = 5000
        k_s = 0.2
        e_s = 200000
        i_s = -1000

        with pytest.raises(NegativeValueError):
            Form5Dot21NominalStiffness(k_c=k_c, e_cd=e_cd, i_c=i_c, k_s=k_s, e_s=e_s, i_s=i_s)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            (
                "complete",
                r"EI = K_{c} \cdot E_{cd} \cdot I_{c} + K_{s} \cdot E_{s} \cdot I_{s} = 0.800 \cdot 30000.000 \cdot 5000.000 + "
                r"0.200 \cdot 200000.000 \cdot 1000.000 = 160000000.000",
            ),
            ("short", r"EI = 160000000.000"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        k_c = 0.8  # -
        e_cd = 30000  # MPa
        i_c = 5000  # mm^4
        k_s = 0.2  # -
        e_s = 200000  # MPa
        i_s = 1000  # mm^4

        # Object to test
        form_5_21_latex = Form5Dot21NominalStiffness(
            k_c=k_c,
            e_cd=e_cd,
            i_c=i_c,
            k_s=k_s,
            e_s=e_s,
            i_s=i_s,
        ).latex()

        actual = {
            "complete": form_5_21_latex.complete,
            "short": form_5_21_latex.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."
